import json
from collections.abc import Generator
from math import isclose
from pathlib import PosixPath
from unittest.mock import MagicMock, patch

import pytest
from dodal.devices.zebra.zebra import RotationDirection

from mx_bluesky.common.parameters.constants import GridscanParamConstants
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    _get_next_instruction,
    _get_pin_type_from_agamemnon_collect_parameters,
    _get_withenergy_parameters_from_agamemnon,
    _get_withvisit_parameters_from_agamemnon,
    _instruction_and_data,
    _PinType,
    _SinglePin,
    compare_params,
    create_parameters_from_agamemnon,
    update_params_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.components import Wait
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect


@pytest.mark.parametrize(
    "num_wells, well_width, buffer, expected_width",
    [
        (3, 500, 0, 1000),
        (6, 50, 100, 450),
        (2, 800, 50, 900),
    ],
)
def test_given_various_pin_formats_then_pin_width_as_expected(
    num_wells, well_width, buffer, expected_width
):
    pin = _PinType(num_wells, well_width, buffer)
    assert pin.full_width == expected_width


def set_up_agamemnon_params(
    loop_type: str | None = None,
    prefix: str | None = None,
    distance: int | None = None,
    wavelength: float | None = None,
):
    return {
        "collection": [{"distance": distance, "wavelength": wavelength}],
        "prefix": prefix,
        "sample": {"loopType": loop_type, "id": 1, "position": 1, "container": 1},
    }


def test_given_no_loop_type_in_parameters_then_single_pin_returned():
    assert (
        _get_pin_type_from_agamemnon_collect_parameters(set_up_agamemnon_params())
        == _SinglePin()
    )


@pytest.mark.parametrize(
    "loop_name, expected_loop",
    [
        ("multipin_6x50+9", _PinType(6, 50, 9)),
        ("multipin_6x25.8+8.6", _PinType(6, 25.8, 8.6)),
        ("multipin_9x31+90", _PinType(9, 31, 90)),
    ],
)
def test_given_multipin_loop_type_in_parameters_then_expected_pin_returned(
    loop_name: str, expected_loop: _PinType
):
    assert (
        _get_pin_type_from_agamemnon_collect_parameters(
            set_up_agamemnon_params(loop_name)
        )
        == expected_loop
    )


@pytest.mark.parametrize(
    "loop_name",
    [
        "nonesense",
        "single_pin_78x89+1",
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
def test_given_completely_unrecognised_loop_type_in_parameters_then_warning_logged_single_pin_returned(
    mock_logger: MagicMock,
    loop_name: str,
):
    assert (
        _get_pin_type_from_agamemnon_collect_parameters(
            set_up_agamemnon_params(loop_name)
        )
        == _SinglePin()
    )
    mock_logger.warning.assert_called_once()


@pytest.mark.parametrize(
    "loop_name",
    [
        "multipin_67x56",
        "multipin_90+4",
        "multipin_8",
        "multipin_6x50+",
        "multipin_6x50+98.",
        "multipin_6x50+.1",
        "multipin_6x.50+98",
        "multipin_6x50+98.1.2",
        "multipin_6x50.5.6+98",
        "multipin_6x50+98..1",
        "multipin_6x.50+.98",
        "multipin_6x+98",
    ],
)
def test_given_unrecognised_multipin_in_parameters_then_warning_logged_single_pin_returned(
    loop_name: str,
):
    with pytest.raises(ValueError) as e:
        _get_pin_type_from_agamemnon_collect_parameters(
            set_up_agamemnon_params(loop_name)
        )
    assert "Expected multipin format" in str(e.value)


def configure_mock_agamemnon(mock_requests: MagicMock, loop_type: str | None):
    mock_requests.get.return_value.content = json.dumps(
        {"collect": set_up_agamemnon_params(loop_type, "", 255, 0.9)}
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_when_get_next_instruction_called_then_expected_agamemnon_url_queried(
    mock_requests: MagicMock,
):
    configure_mock_agamemnon(mock_requests, None)
    _get_next_instruction("i03")
    mock_requests.get.assert_called_once_with(
        "http://agamemnon.diamond.ac.uk/getnextcollect/i03",
        headers={"Accept": "application/json"},
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_returns_an_unexpected_response_then_exception_is_thrown(
    mock_requests: MagicMock,
):
    mock_requests.get.return_value.content = json.dumps({"not_collect": ""})
    with pytest.raises(KeyError) as e:
        create_parameters_from_agamemnon()
    assert "not_collect" in str(e.value)


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_given_agamemnon_returns_multipin_when_get_next_pin_type_from_agamemnon_called_then_multipin_returned(
    mock_requests: MagicMock,
):
    configure_mock_agamemnon(mock_requests, "multipin_6x50+98.1")
    instruction, params = _instruction_and_data(_get_next_instruction("i03"))
    assert _get_pin_type_from_agamemnon_collect_parameters(params) == _PinType(
        6, 50, 98.1
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_leaves_parameters_unchanged_when_agamemnon_fails(
    mock_requests: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    mock_requests.get.side_effect = Exception("Bad")
    old_grid_width = load_centre_collect_params.robot_load_then_centre.grid_width_um
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert params.robot_load_then_centre.grid_width_um == old_grid_width


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.compare_params")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_changes_params_to_single_pin_when_agamemnon_gives_single_pin(
    mock_requests: MagicMock,
    mock_compare_params: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, None)
    load_centre_collect_params.robot_load_then_centre.grid_width_um = 0
    load_centre_collect_params.select_centres.n = 0
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert (
        params.robot_load_then_centre.grid_width_um == GridscanParamConstants.WIDTH_UM
    )
    assert params.select_centres.n == 1
    assert params.multi_rotation_scan.snapshot_omegas_deg


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.compare_params")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_applies_multipin_attribs_given_agamemnon_returns_multipin(
    mock_requests: MagicMock,
    mock_compare_params: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, "multipin_6x50+10")
    params = update_params_from_agamemnon(load_centre_collect_params)
    assert params.robot_load_then_centre.grid_width_um == 270
    assert params.select_centres.n == 6
    assert params.robot_load_then_centre.tip_offset_um == 135
    assert not params.multi_rotation_scan.snapshot_omegas_deg


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_deduces_correct_url_given_set_of_parameters(
    mock_requests: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    update_params_from_agamemnon(load_centre_collect_params)
    mock_requests.get.assert_called_once_with(
        "http://agamemnon.diamond.ac.uk/getnextcollect/i03",
        headers={"Accept": "application/json"},
    )


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_update_params_from_agamemnon_logs_warning_when_exception_occurs(
    mock_requests: MagicMock,
    mock_logger: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, "multipin_unknown")

    update_params_from_agamemnon(load_centre_collect_params)

    mock_logger.warning.assert_called_once()
    assert (
        "Failed to update parameters: Agamemnon loop type of multipin_unknown "
        "not recognised" in mock_logger.warning.mock_calls[0].args[0]
    )


@pytest.mark.parametrize(
    "prefix, expected_visit",
    [
        ["/dls/i03/data/2025/mx23694-130/foo/bar", "mx23694-130"],
        ["/dls/not-i03/data/2021/mx84743-230", "mx84743-230"],
    ],
)
def test_given_valid_prefix_then_correct_visit_is_set(prefix: str, expected_visit: str):
    visit, _ = _get_withvisit_parameters_from_agamemnon(
        set_up_agamemnon_params(None, prefix, None)
    )
    assert visit == expected_visit


@pytest.mark.parametrize(
    "prefix",
    [
        "/not-dls/i03/data/2025/mx23694-130/foo/bar",
        "/dls/i03/not-data/2025/mx23694-130/foo/bar",
        "/foo/bar/i03/data/2025/mx23694-130",
    ],
)
def test_given_invalid_prefix_then_exception_raised(prefix: str):
    with pytest.raises(ValueError) as e:
        _get_withvisit_parameters_from_agamemnon(
            set_up_agamemnon_params(None, prefix, None)
        )

    assert "MX-General root structure" in str(e.value)


def test_no_prefix_raises_exception():
    with pytest.raises(KeyError) as e:
        _get_withvisit_parameters_from_agamemnon({"not_collect": ""})

    assert "Unexpected json from agamemnon" in str(e.value)


@pytest.mark.parametrize(
    "mock_error, mock_log",
    [
        (ValueError(), "Failed to compare parameters: "),
        (Exception(), "Unexpected error occurred. Failed to compare parameters: "),
    ],
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch(
    "mx_bluesky.hyperion.external_interaction.agamemnon._populate_parameters_from_agamemnon"
)
def test_compare_params_logs_exception_if_fails_to_populate_parameters_from_hyperion(
    mock_populate_params,
    mock_logger,
    mock_requests,
    mock_error,
    mock_log,
    load_centre_collect_params: LoadCentreCollect,
):
    configure_mock_agamemnon(mock_requests, None)
    mock_populate_params.side_effect = mock_error
    compare_params(
        load_centre_collect_params,
    )
    assert mock_log in mock_logger.mock_calls[0][1][0]


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch(
    "mx_bluesky.hyperion.external_interaction.agamemnon.create_parameters_from_agamemnon",
    MagicMock(return_value=[]),
)
def test_compare_params_logs_message_if_agamemnon_returns_no_instructions(
    mock_logger: MagicMock, load_centre_collect_params: LoadCentreCollect
):
    compare_params(
        load_centre_collect_params,
    )
    mock_logger.info.assert_called_once_with("Agamemnon returned no instructions")


@pytest.fixture
def agamemnon_response(request) -> Generator[str, None, None]:
    with (
        patch("mx_bluesky.common.parameters.components.os", new=MagicMock()),
        patch(
            "mx_bluesky.hyperion.external_interaction.agamemnon.requests"
        ) as mock_requests,
        open(request.param) as json_file,
    ):
        example_json = json_file.read()
        mock_requests.get.return_value.content = example_json
        yield example_json


@pytest.mark.parametrize(
    "agamemnon_response",
    [
        "tests/test_data/agamemnon/example_native.json",
        "tests/test_data/agamemnon/example_collect_multipin.json",
    ],
    indirect=True,
)
@patch("mx_bluesky.hyperion.external_interaction.agamemnon.LOGGER")
@patch("mx_bluesky.hyperion.parameters.rotation.os", new=MagicMock())
@patch("dodal.devices.detector.detector.Path", new=MagicMock())
@patch("dodal.utils.os", new=MagicMock())
def test_compare_params_causes_no_warning_when_compared_to_gda_params(
    mock_logger: MagicMock,
    agamemnon_response: str,
    load_centre_collect_params: LoadCentreCollect,
):
    compare_params(load_centre_collect_params)
    mock_logger.warning.assert_not_called()


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_contains_expected_data(agamemnon_response):
    hyperion_params_list = create_parameters_from_agamemnon()
    for hyperion_params in hyperion_params_list:
        assert isinstance(hyperion_params, LoadCentreCollect)
        assert hyperion_params.visit == "mx34598-77"
        assert isclose(hyperion_params.detector_distance_mm, 237.017, abs_tol=1e-3)  # type: ignore
        assert hyperion_params.sample_id == 6501159
        assert hyperion_params.sample_puck == 5
        assert hyperion_params.sample_pin == 4
        assert str(hyperion_params.parameter_model_version) == "5.3.0"
        assert hyperion_params.select_centres.n == 1


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_contains_expected_robot_load_then_centre_data(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    load_centre_collect_list = [
        p for p in hyperion_params_list if isinstance(p, LoadCentreCollect)
    ]
    assert len(hyperion_params_list) == len(load_centre_collect_list) == 2

    assert load_centre_collect_list[0].robot_load_then_centre.chi_start_deg == 0.0
    assert load_centre_collect_list[1].robot_load_then_centre.chi_start_deg == 30.0
    for robot_load_params in [
        params.robot_load_then_centre for params in load_centre_collect_list
    ]:
        assert robot_load_params.visit == "mx34598-77"
        assert isclose(robot_load_params.detector_distance_mm, 237.017, abs_tol=1e-3)  # type: ignore
        assert robot_load_params.sample_id == 6501159
        assert robot_load_params.sample_puck == 5
        assert robot_load_params.sample_pin == 4
        assert robot_load_params.demand_energy_ev == 12700.045934258673
        assert robot_load_params.omega_start_deg == 0.0
        assert robot_load_params.transmission_frac == 1.0
        assert robot_load_params.tip_offset_um == 300.0
        assert robot_load_params.grid_width_um == 600.0
        assert str(robot_load_params.parameter_model_version) == "5.3.0"
        assert (
            robot_load_params.storage_directory
            == "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242/xraycentring"
        )
        assert robot_load_params.file_name == "CBLBA-x00242"
        assert robot_load_params.snapshot_directory == PosixPath(
            "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242/xraycentring/snapshots"
        )


@patch("mx_bluesky.hyperion.parameters.rotation.os", new=MagicMock())
@patch("dodal.devices.detector.detector.Path", new=MagicMock())
@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_contains_expected_rotation_data(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    assert len(hyperion_params_list) == 2
    for hyperion_params in hyperion_params_list:
        assert isinstance(hyperion_params, LoadCentreCollect)
        rotation_params = hyperion_params.multi_rotation_scan
        assert rotation_params.visit == "mx34598-77"
        assert isclose(rotation_params.detector_distance_mm, 237.017, abs_tol=1e-3)  # type: ignore
        assert rotation_params.detector_params.omega_start == 0.0
        assert rotation_params.detector_params.exposure_time_s == 0.003
        assert rotation_params.detector_params.num_images_per_trigger == 3600
        assert rotation_params.num_images == 3600
        assert rotation_params.transmission_frac == 0.1426315789473684
        assert rotation_params.comment == "Complete_P1_sweep1 "
        assert rotation_params.ispyb_experiment_type == "OSC"

        assert rotation_params.demand_energy_ev == 12700.045934258673
        assert str(rotation_params.parameter_model_version) == "5.3.0"
        assert (
            rotation_params.storage_directory
            == "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242"
        )
        assert rotation_params.file_name == "CBLBA-x00242"
        assert rotation_params.snapshot_directory == PosixPath(
            "/dls/i03/data/2025/mx34598-77/auto/CBLBA/CBLBA-x00242/snapshots"
        )

    individual_scans = list(
        hyperion_params_list[0].multi_rotation_scan.single_rotation_scans  # type: ignore
    ) + list(
        hyperion_params_list[1].multi_rotation_scan.single_rotation_scans  # type: ignore
    )
    assert len(individual_scans) == 2
    assert individual_scans[0].scan_points["omega"][1] == 0.1
    assert individual_scans[0].phi_start_deg == 0.0
    assert individual_scans[0].chi_start_deg == 0.0
    assert individual_scans[0].rotation_direction == RotationDirection.POSITIVE
    assert individual_scans[1].scan_points["omega"][1] == 0.1
    assert individual_scans[1].phi_start_deg == 0.0
    assert individual_scans[1].chi_start_deg == 30.0
    assert individual_scans[1].rotation_direction == RotationDirection.POSITIVE


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_collect_multipin.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_populates_multipin_parameters_from_agamemnon(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    for hyperion_params in hyperion_params_list:
        assert hyperion_params.select_centres.n == 6  # type: ignore


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_creates_multiple_load_centre_collect_for_native_collection(
    agamemnon_response,
):
    hyperion_params_list = create_parameters_from_agamemnon()
    assert len(hyperion_params_list) == 2
    assert (
        sum(
            [
                len(hyperion_params.multi_rotation_scan.rotation_scans)  # type: ignore
                for hyperion_params in hyperion_params_list
            ]
        )
        == 2
    )


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_native.json"],
    indirect=True,
)
def test_get_withenergy_parameters_from_agamemnon(agamemnon_response):
    _, agamemnon_params = _instruction_and_data(_get_next_instruction("i03"))
    demand_energy_ev = _get_withenergy_parameters_from_agamemnon(agamemnon_params)
    assert demand_energy_ev["demand_energy_ev"] == 12700.045934258673


def test_get_withenergy_parameters_from_agamemnon_when_no_wavelength():
    agamemnon_params = {}
    demand_energy_ev = _get_withenergy_parameters_from_agamemnon(agamemnon_params)
    assert demand_energy_ev["demand_energy_ev"] is None


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_create_parameters_from_agamemnon_returns_empty_list_if_collect_instruction_is_empty(
    mock_requests,
):
    mock_requests.get.return_value.content = json.dumps({"collect": {}})
    params = create_parameters_from_agamemnon()
    assert params == []


@patch("mx_bluesky.hyperion.external_interaction.agamemnon.requests")
def test_create_parameters_from_agamemnon_returns_empty_list_if_no_instruction(
    mock_requests,
):
    mock_requests.get.return_value.content = json.dumps({})
    params = create_parameters_from_agamemnon()
    assert params == []


@pytest.mark.parametrize(
    "agamemnon_response",
    ["tests/test_data/agamemnon/example_collect_multipin.json"],
    indirect=True,
)
def test_create_parameters_from_agamemnon_does_not_return_none_if_queue_is_not_empty(
    agamemnon_response,
):
    params = create_parameters_from_agamemnon()
    assert params is not None


@pytest.mark.parametrize(
    "agamemnon_response", ["tests/test_data/agamemnon/example_wait.json"], indirect=True
)
def test_create_parameters_from_agamemnon_creates_wait(agamemnon_response):
    params = create_parameters_from_agamemnon()
    assert len(params) == 1
    assert isinstance(params[0], Wait)
    assert params[0].duration_s == 12.34



import os
import re
from collections.abc import Generator
from decimal import Decimal
from functools import partial
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from aiohttp import ClientResponse
from dodal.beamlines import i03
from dodal.devices.oav.oav_parameters import OAVConfigBeamCentre
from ophyd_async.core import AsyncStatus
from ophyd_async.testing import set_mock_value
from PIL import Image

# Map all the case-sensitive column names from their normalised versions
DATA_COLLECTION_COLUMN_MAP = {
    s.lower(): s
    for s in [
        "dataCollectionId",
        "BLSAMPLEID",
        "SESSIONID",
        "experimenttype",
        "dataCollectionNumber",
        "startTime",
        "endTime",
        "runStatus",
        "axisStart",
        "axisEnd",
        "axisRange",
        "overlap",
        "numberOfImages",
        "startImageNumber",
        "numberOfPasses",
        "exposureTime",
        "imageDirectory",
        "imagePrefix",
        "imageSuffix",
        "imageContainerSubPath",
        "fileTemplate",
        "wavelength",
        "resolution",
        "detectorDistance",
        "xBeam",
        "yBeam",
        "comments",
        "printableForReport",
        "CRYSTALCLASS",
        "slitGapVertical",
        "slitGapHorizontal",
        "transmission",
        "synchrotronMode",
        "xtalSnapshotFullPath1",
        "xtalSnapshotFullPath2",
        "xtalSnapshotFullPath3",
        "xtalSnapshotFullPath4",
        "rotationAxis",
        "phiStart",
        "kappaStart",
        "omegaStart",
        "chiStart",
        "resolutionAtCorner",
        "detector2Theta",
        "DETECTORMODE",
        "undulatorGap1",
        "undulatorGap2",
        "undulatorGap3",
        "beamSizeAtSampleX",
        "beamSizeAtSampleY",
        "centeringMethod",
        "averageTemperature",
        "ACTUALSAMPLEBARCODE",
        "ACTUALSAMPLESLOTINCONTAINER",
        "ACTUALCONTAINERBARCODE",
        "ACTUALCONTAINERSLOTINSC",
        "actualCenteringPosition",
        "beamShape",
        "dataCollectionGroupId",
        "POSITIONID",
        "detectorId",
        "FOCALSPOTSIZEATSAMPLEX",
        "POLARISATION",
        "FOCALSPOTSIZEATSAMPLEY",
        "APERTUREID",
        "screeningOrigId",
        "flux",
        "strategySubWedgeOrigId",
        "blSubSampleId",
        "processedDataFile",
        "datFullPath",
        "magnification",
        "totalAbsorbedDose",
        "binning",
        "particleDiameter",
        "boxSize",
        "minResolution",
        "minDefocus",
        "maxDefocus",
        "defocusStepSize",
        "amountAstigmatism",
        "extractSize",
        "bgRadius",
        "voltage",
        "objAperture",
        "c1aperture",
        "c2aperture",
        "c3aperture",
        "c1lens",
        "c2lens",
        "c3lens",
        "startPositionId",
        "endPositionId",
        "flux",
        "bestWilsonPlotPath",
        "totalExposedDose",
        "nominalMagnification",
        "nominalDefocus",
        "imageSizeX",
        "imageSizeY",
        "pixelSizeOnImage",
        "phasePlate",
        "dataCollectionPlanId",
    ]
}


@pytest.fixture(autouse=True, scope="session")
def ispyb_config_path() -> Generator[str, Any, Any]:
    ispyb_config_path = os.environ.get(
        "ISPYB_CONFIG_PATH", "/dls_sw/dasc/mariadb/credentials/ispyb-hyperion-dev.cfg"
    )
    with patch.dict(os.environ, {"ISPYB_CONFIG_PATH": ispyb_config_path}):
        yield ispyb_config_path


@pytest.fixture
def undulator_for_system_test(undulator):
    set_mock_value(undulator.current_gap, 1.11)
    return undulator


@pytest.fixture
def next_oav_system_test_image():
    return MagicMock(
        return_value="tests/test_data/test_images/generate_snapshot_input.png"
    )


@pytest.fixture
def oav_for_system_test(test_config_files, next_oav_system_test_image):
    parameters = OAVConfigBeamCentre(
        test_config_files["zoom_params_file"], test_config_files["display_config"]
    )
    oav = i03.oav(connect_immediately=True, mock=True, params=parameters)
    set_mock_value(oav.cam.array_size_x, 1024)
    set_mock_value(oav.cam.array_size_y, 768)

    # Grid snapshots
    set_mock_value(oav.grid_snapshot.x_size, 1024)
    set_mock_value(oav.grid_snapshot.y_size, 768)
    set_mock_value(oav.grid_snapshot.top_left_x, 50)
    set_mock_value(oav.grid_snapshot.top_left_y, 100)
    size_in_pixels = 0.1 * 1000 / 1.25
    set_mock_value(oav.grid_snapshot.box_width, size_in_pixels)

    # Rotation snapshots
    @AsyncStatus.wrap
    async def trigger_with_test_image(self):
        with Image.open(next_oav_system_test_image()) as image:
            await self.post_processing(image)

    oav.snapshot.trigger = MagicMock(
        side_effect=partial(trigger_with_test_image, oav.snapshot)
    )
    oav.grid_snapshot.trigger = MagicMock(
        side_effect=partial(trigger_with_test_image, oav.grid_snapshot)
    )

    empty_response = AsyncMock(spec=ClientResponse)
    empty_response.read.return_value = b""

    with (
        patch(
            "dodal.devices.areadetector.plugins.MJPG.ClientSession.get", autospec=True
        ) as mock_get,
    ):
        mock_get.return_value.__aenter__.return_value = empty_response
        set_mock_value(oav.zoom_controller.level, "1.0x")
        zoom_levels_list = ["1.0x", "3.0x", "5.0x", "7.5x", "10.0x", "15.0x"]
        oav.zoom_controller._get_allowed_zoom_levels = AsyncMock(
            return_value=zoom_levels_list
        )
        yield oav


def compare_actual_and_expected(
    id, expected_values, fetch_datacollection_attribute, column_map: dict | None = None
):
    results = "\n"
    for k, v in expected_values.items():
        actual = fetch_datacollection_attribute(
            id, column_map[k.lower()] if column_map else k
        )
        if isinstance(actual, Decimal):
            actual = float(actual)
        if isinstance(v, float):
            actual_v = actual == pytest.approx(v)
        elif isinstance(v, str) and v.startswith("regex:"):
            actual_v = re.match(v.removeprefix("regex:"), str(actual))  # type: ignore
        else:
            actual_v = actual == v
        if not actual_v:
            results += f"expected {k} {v} == {actual}\n"
    assert results == "\n", results


def compare_comment(
    fetch_datacollection_attribute, data_collection_id, expected_comment
):
    actual_comment = fetch_datacollection_attribute(
        data_collection_id, DATA_COLLECTION_COLUMN_MAP["comments"]
    )
    match = re.search(" Zocalo processing took", actual_comment)
    truncated_comment = actual_comment[: match.start()] if match else actual_comment
    assert truncated_comment == expected_comment



from unittest.mock import MagicMock

import pytest
from flask import Flask
from flask.testing import FlaskClient

from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.hyperion.plan_runner import PlanRunner
from mx_bluesky.hyperion.plan_runner_api import (
    create_app_for_udc,
)


@pytest.fixture()
def mock_runner():
    return MagicMock(spec=PlanRunner)


@pytest.fixture()
def app_under_test(mock_runner):
    app = create_app_for_udc(mock_runner)
    yield app


@pytest.fixture()
def client(app_under_test: Flask) -> FlaskClient:
    return app_under_test.test_client()


def test_plan_runner_api_fetch_status(app_under_test, client, mock_runner):
    mock_runner.current_status = Status.BUSY
    response = client.get("/status")
    assert response.status_code == 200
    assert response.content_type == "application/json"
    assert response.json["status"] == Status.BUSY.value



from collections.abc import Iterator
from os import environ
from unittest.mock import patch

import pytest

environ["HYPERION_TEST_MODE"] = "true"


def pytest_addoption(parser):
    parser.addoption(
        "--logging",
        action="store_true",
        default=False,
        help="Log during all tests (not just those that are testing logging logic)",
    )


@pytest.fixture(scope="session", autouse=True)
def default_session_fixture() -> Iterator[None]:
    print("Patching bluesky 0MQ Publisher in __main__ for the whole session")
    with patch("mx_bluesky.hyperion.runner.Publisher"):
        yield



from functools import cache

from mx_bluesky.common.external_interaction.config_server import MXConfigClient
from mx_bluesky.hyperion.parameters.constants import (
    HyperionFeatureSetting,
    HyperionFeatureSettingources,
)


@cache
def get_hyperion_config_client() -> MXConfigClient[HyperionFeatureSetting]:
    return MXConfigClient(
        feature_sources=HyperionFeatureSettingources,
        feature_dc=HyperionFeatureSetting,
        url="https://daq-config.diamond.ac.uk",
    )



from __future__ import annotations

from blueapi.core import BlueskyContext
from bluesky.utils import MsgGenerator
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)

from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    grid_detect_then_xray_centre,
)
from mx_bluesky.common.parameters.constants import OavConstants, PlanNameConstants
from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_decorator,
)
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.hyperion.experiment_plans.hyperion_flyscan_xray_centre_plan import (
    construct_hyperion_specific_features,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)


def create_devices(
    context: BlueskyContext,
) -> HyperionGridDetectThenXRayCentreComposite:
    return device_composite_from_context(
        context, HyperionGridDetectThenXRayCentreComposite
    )


def hyperion_grid_detect_then_xray_centre(
    composite: HyperionGridDetectThenXRayCentreComposite,
    parameters: GridScanWithEdgeDetect,
    oav_config: str = OavConstants.OAV_CONFIG_JSON,
) -> MsgGenerator:
    """
    A plan which combines the collection of snapshots from the OAV and the determination
    of the grid dimensions to use for the following grid scan.
    """

    @verify_undulator_gap_before_run_decorator(composite)
    @transmission_and_xbpm_feedback_for_collection_decorator(
        composite, parameters.transmission_frac, PlanNameConstants.GRIDSCAN_OUTER
    )
    def plan_to_perform():
        yield from grid_detect_then_xray_centre(
            composite=composite,
            parameters=parameters,
            xrc_params_type=HyperionSpecifiedThreeDGridScan,
            construct_beamline_specific=construct_hyperion_specific_features,
            oav_config=oav_config,
        )

    yield from plan_to_perform()



import math

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.beamlines import aithre
from dodal.devices.aithre_lasershaping.goniometer import Goniometer
from dodal.testing import patch_motor
from ophyd_async.core import init_devices
from ophyd_async.epics.motor import Motor
from ophyd_async.testing import set_mock_value

from mx_bluesky.beamlines.aithre_lasershaping import (
    change_goniometer_turn_speed,
    go_to_furthest_maximum,
    jog_sample,
    rotate_goniometer_relative,
)
from mx_bluesky.beamlines.aithre_lasershaping.goniometer_controls import JogDirection


@pytest.fixture
def goniometer(RE: RunEngine) -> Goniometer:
    with init_devices(mock=True):
        gonio = aithre.goniometer(connect_immediately=True, mock=True)

    patch_motor(gonio.omega)
    patch_motor(gonio.x)
    patch_motor(gonio.z)
    patch_motor(gonio.sampy)
    patch_motor(gonio.sampz)
    return gonio


def test_goniometer_relative_rotation(
    sim_run_engine: RunEngineSimulator, goniometer: Goniometer
):
    msgs = sim_run_engine.simulate_plan(rotate_goniometer_relative(15, goniometer))
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "goniometer-omega"
        and msg.args[0] == 15,
    )


def test_change_goniometer_turn_speed(
    sim_run_engine: RunEngineSimulator, goniometer: Goniometer
):
    msgs = sim_run_engine.simulate_plan(change_goniometer_turn_speed(40, goniometer))
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "goniometer-omega-velocity"
        and msg.args[0] == 40,
    )


@pytest.mark.parametrize(
    "initial_position, expected_set_value", [(-1, 3600), (0, 3600), (3600, -3600)]
)
async def test_go_to_furthest_maximum_real_run_engine(
    goniometer: Goniometer,
    initial_position: float,
    expected_set_value: float,
    RE: RunEngine,
):
    set_mock_value(goniometer.omega.user_readback, initial_position)

    RE(go_to_furthest_maximum(goniometer))

    assert await goniometer.omega.user_setpoint.get_value() == expected_set_value


@pytest.mark.parametrize(
    "directions, axis",
    [
        ((JogDirection.RIGHT, JogDirection.LEFT), "x"),
        ((JogDirection.ZPLUS, JogDirection.ZMINUS), "z"),
    ],
)
async def test_jog_sample_x_z(RE: RunEngine, goniometer: Goniometer, directions, axis):
    goniometer_axis: Motor = getattr(goniometer, axis)

    RE(jog_sample(directions[0], 0.05, goniometer))
    assert await goniometer_axis.user_readback.get_value() == 0.05

    RE(jog_sample(directions[1], 0.05, goniometer))
    assert await goniometer_axis.user_readback.get_value() == 0


async def test_jog_sample_up_down(RE: RunEngine, goniometer: Goniometer):
    set_mock_value(goniometer.omega.user_readback, 60)

    RE(jog_sample(JogDirection.UP, 1, goniometer))
    assert await goniometer.sampz.user_readback.get_value() == pytest.approx(0.5)
    assert await goniometer.sampy.user_readback.get_value() == pytest.approx(
        math.sqrt(3) / 2
    )

    RE(jog_sample(JogDirection.UP, 1, goniometer))
    assert await goniometer.sampz.user_readback.get_value() == pytest.approx(1)
    assert await goniometer.sampy.user_readback.get_value() == pytest.approx(
        math.sqrt(3)
    )

    RE(jog_sample(JogDirection.DOWN, 1, goniometer))
    assert await goniometer.sampz.user_readback.get_value() == pytest.approx(0.5)
    assert await goniometer.sampy.user_readback.get_value() == pytest.approx(
        math.sqrt(3) / 2
    )



from __future__ import annotations

import os
from collections.abc import Callable, Generator, Iterable
from contextlib import nullcontext
from functools import partial
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

import bluesky.plan_stubs as bps
import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.synchrotron import SynchrotronMode
from ispyb.sqlalchemy import BLSample
from ophyd.sim import NullStatus
from ophyd_async.core import AsyncStatus
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    detect_grid_and_do_gridscan,
)
from mx_bluesky.common.external_interaction.callbacks.common.grid_detection_callback import (
    GridParamUpdate,
)
from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    get_proposal_and_session_from_visit_string,
)
from mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback import (
    SampleHandlingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.parameters.components import TopNByMaxCountForEachSampleSelection
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundException,
    WarningException,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
    load_centre_collect_full,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import _move_and_rotation
from mx_bluesky.hyperion.external_interaction.callbacks.robot_actions.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback import (
    BeamDrawingCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.gridscan import GridCommonWithHyperionDetectorParams
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect

from ....conftest import (
    TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX,
    TEST_RESULT_IN_BOUNDS_TOP_LEFT_GRID_CORNER,
    TEST_RESULT_MEDIUM,
    TEST_RESULT_OUT_OF_BOUNDS_BB,
    TEST_RESULT_OUT_OF_BOUNDS_COM,
    TEST_RESULT_SMALL,
    SimConstants,
    assert_images_pixelwise_equal,
    raw_params_from_file,
    replace_all_tmp_paths,
)
from ...conftest import (
    DATA_COLLECTION_COLUMN_MAP,
    compare_actual_and_expected,
    compare_comment,
)

SNAPSHOT_GENERATION_ZOCALO_RESULT = [
    {
        "centre_of_mass": [7.25, 12.2, 5.38],
        "max_voxel": [7, 12, 5],
        "max_count": 50000,
        "n_voxels": 35,
        "total_count": 100000,
        "bounding_box": [[1, 2, 3], [3, 4, 4]],
        "sample_id": SimConstants.ST_SAMPLE_ID,
    }
]


@pytest.fixture
def load_centre_collect_params(tmp_path):
    json_dict = raw_params_from_file(
        "tests/test_data/parameter_json_files/example_load_centre_collect_params.json",
        tmp_path,
    )
    json_dict["visit"] = SimConstants.ST_VISIT
    json_dict["sample_id"] = SimConstants.ST_SAMPLE_ID
    return LoadCentreCollect(**json_dict)


@pytest.fixture
def load_centre_collect_msp_params(load_centre_collect_params):
    load_centre_collect_params.select_centres = TopNByMaxCountForEachSampleSelection(
        n=5
    )
    load_centre_collect_params.sample_id = SimConstants.ST_MSP_SAMPLE_IDS[0]
    load_centre_collect_params.robot_load_then_centre.sample_id = (
        load_centre_collect_params.sample_id
    )
    return load_centre_collect_params


@pytest.fixture
def load_centre_collect_composite(
    grid_detect_then_xray_centre_composite,
    beamstop_phase1,
    composite_for_rotation_scan,
    thawer,
    vfm,
    mirror_voltages,
    undulator_dcm,
    webcam,
    lower_gonio,
    baton,
):
    composite = LoadCentreCollectComposite(
        aperture_scatterguard=composite_for_rotation_scan.aperture_scatterguard,
        attenuator=composite_for_rotation_scan.attenuator,
        backlight=composite_for_rotation_scan.backlight,
        baton=baton,
        beamstop=beamstop_phase1,
        dcm=composite_for_rotation_scan.dcm,
        detector_motion=composite_for_rotation_scan.detector_motion,
        eiger=grid_detect_then_xray_centre_composite.eiger,
        flux=composite_for_rotation_scan.flux,
        robot=composite_for_rotation_scan.robot,
        smargon=composite_for_rotation_scan.smargon,
        undulator=composite_for_rotation_scan.undulator,
        synchrotron=composite_for_rotation_scan.synchrotron,
        s4_slit_gaps=composite_for_rotation_scan.s4_slit_gaps,
        sample_shutter=composite_for_rotation_scan.sample_shutter,
        zebra=grid_detect_then_xray_centre_composite.zebra,
        oav=grid_detect_then_xray_centre_composite.oav,
        xbpm_feedback=composite_for_rotation_scan.xbpm_feedback,
        zebra_fast_grid_scan=grid_detect_then_xray_centre_composite.zebra_fast_grid_scan,
        pin_tip_detection=grid_detect_then_xray_centre_composite.pin_tip_detection,
        zocalo=grid_detect_then_xray_centre_composite.zocalo,
        panda=grid_detect_then_xray_centre_composite.panda,
        panda_fast_grid_scan=grid_detect_then_xray_centre_composite.panda_fast_grid_scan,
        thawer=thawer,
        vfm=vfm,
        mirror_voltages=mirror_voltages,
        undulator_dcm=undulator_dcm,
        webcam=webcam,
        lower_gonio=lower_gonio,
    )

    set_mock_value(composite.dcm.bragg_in_degrees.user_readback, 5)

    yield composite


@pytest.fixture
def robot_load_cb() -> RobotLoadISPyBCallback:
    robot_load_cb = RobotLoadISPyBCallback()
    robot_load_cb.expeye.start_robot_action = MagicMock(return_value=1234)
    robot_load_cb.expeye.end_robot_action = MagicMock()
    robot_load_cb.expeye.update_robot_action = MagicMock()
    return robot_load_cb


GRID_DC_1_EXPECTED_VALUES = {
    "BLSAMPLEID": SimConstants.ST_SAMPLE_ID,
    "detectorid": 78,
    "axisstart": 90.0,
    "axisrange": 0,
    "axisend": 90,
    "focalspotsizeatsamplex": 0.02,
    "focalspotsizeatsampley": 0.02,
    "slitgapvertical": 0.234,
    "slitgaphorizontal": 0.123,
    "beamsizeatsamplex": 0.02,
    "beamsizeatsampley": 0.02,
    "transmission": 100,
    "datacollectionnumber": 1,
    "detectordistance": 255.0,
    "exposuretime": 0.002,
    "imagedirectory": "{tmp_data}/123457/xraycentring/",
    "imageprefix": "robot_load_centring_file",
    "imagesuffix": "h5",
    "numberofpasses": 1,
    "overlap": 0,
    "omegastart": 90,
    "startimagenumber": 1,
    "wavelength": 1.11697,
    "xbeam": 75.6027,
    "ybeam": 79.4935,
    "xtalsnapshotfullpath1": "{tmp_data}/123457/xraycentring/snapshots/robot_load_centring_file_1_90_grid_overlay.png",
    "xtalsnapshotfullpath2": "{tmp_data}/123457/xraycentring/snapshots"
    "/robot_load_centring_file_1_90_outer_overlay.png",
    "xtalsnapshotfullpath3": "{tmp_data}/123457/xraycentring/snapshots/robot_load_centring_file_1_90.png",
    "synchrotronmode": "User",
    "undulatorgap1": 1.11,
    "filetemplate": "robot_load_centring_file_1_master.h5",
    "numberofimages": 180,
}

GRID_DC_2_EXPECTED_VALUES = GRID_DC_1_EXPECTED_VALUES | {
    "axisstart": 0,
    "axisend": 0,
    "omegastart": 0,
    "datacollectionnumber": 2,
    "filetemplate": "robot_load_centring_file_2_master.h5",
    "numberofimages": 180,
    "xtalsnapshotfullpath1": "{tmp_data}/123457/xraycentring/snapshots"
    "/robot_load_centring_file_1_0_grid_overlay.png",
    "xtalsnapshotfullpath2": "{tmp_data}/123457/xraycentring/snapshots"
    "/robot_load_centring_file_1_0_outer_overlay.png",
    "xtalsnapshotfullpath3": "{tmp_data}/123457/xraycentring/snapshots"
    "/robot_load_centring_file_1_0.png",
}

ROTATION_DC_EXPECTED_VALUES = {
    "axisStart": 10,
    "axisEnd": -350,
    # "chiStart": 0, mx-bluesky 325
    "wavelength": 1.11697,
    "beamSizeAtSampleX": 0.02,
    "beamSizeAtSampleY": 0.02,
    "exposureTime": 0.004,
    "undulatorGap1": 1.11,
    "synchrotronMode": SynchrotronMode.USER.value,
    "slitGapHorizontal": 0.123,
    "slitGapVertical": 0.234,
    "xtalSnapshotFullPath1": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_0_with_beam_centre\\.png",
    "xtalSnapshotFullPath2": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_90_with_beam_centre\\.png",
    "xtalSnapshotFullPath3": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_180_with_beam_centre\\.png",
    "xtalSnapshotFullPath4": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_270_with_beam_centre\\.png",
}

ROTATION_DC_2_EXPECTED_VALUES = ROTATION_DC_EXPECTED_VALUES | {
    "axisStart": -350,
    "axisEnd": 10,
    "xtalSnapshotFullPath1": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_0_with_beam_centre\\.png",
    "xtalSnapshotFullPath2": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_90_with_beam_centre\\.png",
    "xtalSnapshotFullPath3": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_180_with_beam_centre\\.png",
    "xtalSnapshotFullPath4": "regex:{tmp_data}/123457/snapshots/\\d{"
    "8}_oav_snapshot_270_with_beam_centre\\.png",
}


@pytest.fixture
def composite_with_no_diffraction(
    load_centre_collect_composite: LoadCentreCollectComposite,
) -> Generator[LoadCentreCollectComposite, Any, None]:
    zocalo = load_centre_collect_composite.zocalo

    @AsyncStatus.wrap
    async def mock_zocalo_complete():
        await zocalo._put_results([], {"dcid": 0, "dcgid": 0})

    with patch.object(zocalo, "trigger", side_effect=mock_zocalo_complete):
        yield load_centre_collect_composite


@pytest.mark.system_test
def test_execute_load_centre_collect_full(
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollectiongroup_attribute: Callable[..., Any],
    fetch_datacollection_ids_for_group_id: Callable[..., Any],
    fetch_blsample: Callable[[int], BLSample],
    tmp_path,
    robot_load_cb: RobotLoadISPyBCallback,
):
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    snapshot_cb = BeamDrawingCallback(emit=ispyb_rotation_cb)
    set_mock_value(
        load_centre_collect_composite.undulator_dcm.undulator_ref().current_gap, 1.11
    )
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(snapshot_cb)
    RE.subscribe(robot_load_cb)
    RE(
        load_centre_collect_full(
            load_centre_collect_composite,
            load_centre_collect_params,
            oav_parameters_for_rotation,
        )
    )

    expected_proposal, expected_visit = get_proposal_and_session_from_visit_string(
        load_centre_collect_params.visit
    )
    expected_sample_id = load_centre_collect_params.sample_id
    robot_load_cb.expeye.start_robot_action.assert_called_once_with(  # type: ignore
        "LOAD", expected_proposal, expected_visit, expected_sample_id
    )
    # TODO re-enable this https://github.com/DiamondLightSource/mx-bluesky/issues/690
    # robot_load_cb.expeye.update_barcode_and_snapshots.assert_called_once_with(
    #     1234,
    #     "BARCODE",
    #     "{tmp_data}/123457/xraycentring/snapshots/160705_webcam_after_load.png",
    #     "/tmp/snapshot1.png",
    # )
    robot_load_cb.expeye.end_robot_action.assert_called_once_with(1234, "success", "OK")  # type: ignore

    # Compare gridscan collection
    compare_actual_and_expected(
        ispyb_gridscan_cb.ispyb_ids.data_collection_group_id,
        {"experimentType": "Mesh3D", "blSampleId": expected_sample_id},
        fetch_datacollectiongroup_attribute,
    )
    compare_actual_and_expected(
        ispyb_gridscan_cb.ispyb_ids.data_collection_ids[0],
        replace_all_tmp_paths(GRID_DC_1_EXPECTED_VALUES, tmp_path),
        fetch_datacollection_attribute,
        DATA_COLLECTION_COLUMN_MAP,
    )
    compare_actual_and_expected(
        ispyb_gridscan_cb.ispyb_ids.data_collection_ids[1],
        replace_all_tmp_paths(GRID_DC_2_EXPECTED_VALUES, tmp_path),
        fetch_datacollection_attribute,
        DATA_COLLECTION_COLUMN_MAP,
    )

    compare_comment(
        fetch_datacollection_attribute,
        ispyb_gridscan_cb.ispyb_ids.data_collection_ids[0],
        "MX-Bluesky: Xray centring 1 - Diffraction grid scan of 30 by 6 "
        "images in 20.0 um by 20.0 um steps. Top left (px): [130,130], "
        "bottom right (px): [874,278]. Aperture: Small. ",
    )
    compare_comment(
        fetch_datacollection_attribute,
        ispyb_gridscan_cb.ispyb_ids.data_collection_ids[1],
        "MX-Bluesky: Xray centring 2 - Diffraction grid scan of 30 by 6 "
        "images in 20.0 um by 20.0 um steps. Top left (px): [130,130], "
        "bottom right (px): [874,278]. Aperture: Small. ",
    )

    rotation_dcg_id = ispyb_rotation_cb.ispyb_ids.data_collection_group_id
    rotation_dc_ids = fetch_datacollection_ids_for_group_id(rotation_dcg_id)
    compare_actual_and_expected(
        rotation_dcg_id,
        {"experimentType": "SAD", "blSampleId": expected_sample_id},
        fetch_datacollectiongroup_attribute,
    )
    compare_actual_and_expected(
        rotation_dc_ids[0],
        replace_all_tmp_paths(ROTATION_DC_EXPECTED_VALUES, tmp_path),
        fetch_datacollection_attribute,
    )
    compare_actual_and_expected(
        rotation_dc_ids[1],
        replace_all_tmp_paths(ROTATION_DC_2_EXPECTED_VALUES, tmp_path),
        fetch_datacollection_attribute,
    )

    compare_comment(
        fetch_datacollection_attribute,
        ispyb_rotation_cb.ispyb_ids.data_collection_ids[0],
        "Hyperion Rotation Scan -  Sample position (µm): (-2309, -591, -571) Aperture: "
        "Small. ",
    )
    assert fetch_blsample(expected_sample_id).blSampleStatus == "LOADED"  # type: ignore


@pytest.mark.system_test
def test_load_centre_collect_updates_bl_sample_status_robot_load_fail(
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_blsample: Callable[..., Any],
):
    robot_load_cb = RobotLoadISPyBCallback()
    sample_handling_cb = SampleHandlingCallback()
    RE.subscribe(robot_load_cb)
    RE.subscribe(sample_handling_cb)

    with (
        patch(
            "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.wait_for_smargon_not_disabled",
            side_effect=TimeoutError("Simulated timeout"),
        ),
        pytest.raises(TimeoutError, match="Simulated timeout"),
    ):
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_params,
                oav_parameters_for_rotation,
            )
        )

    assert (
        fetch_blsample(load_centre_collect_params.sample_id).blSampleStatus
        == "ERROR - beamline"
    )


@pytest.mark.system_test
def test_load_centre_collect_updates_bl_sample_status_pin_tip_detection_fail(
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    pin_tip_no_pin_found: PinTipDetection,
    RE: RunEngine,
    fetch_blsample: Callable[..., Any],
):
    robot_load_cb = RobotLoadISPyBCallback()
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    sample_handling_cb = SampleHandlingCallback()
    RE.subscribe(robot_load_cb)
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(sample_handling_cb)

    with pytest.raises(
        WarningException, match="Pin tip centring failed - pin too long/short.*"
    ):
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_params,
                oav_parameters_for_rotation,
            )
        )

    assert (
        fetch_blsample(load_centre_collect_params.sample_id).blSampleStatus
        == "ERROR - sample"
    )


@pytest.mark.system_test
def test_load_centre_collect_updates_bl_sample_status_grid_detection_fail_tip_not_found(
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_blsample: Callable[..., Any],
):
    robot_load_cb = RobotLoadISPyBCallback()
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    sample_handling_cb = SampleHandlingCallback()
    RE.subscribe(robot_load_cb)
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(sample_handling_cb)

    descriptor = None

    def wait_for_first_oav_grid(name: str, doc: dict):
        nonlocal descriptor
        if (
            name == "descriptor"
            and doc["name"] == CONST.DESCRIPTORS.OAV_GRID_SNAPSHOT_TRIGGERED
        ):
            descriptor = doc["uid"]
        if name == "event" and doc["descriptor"] == descriptor:
            # Trigger a fail to find the pin at 2nd grid detect
            set_mock_value(
                load_centre_collect_composite.pin_tip_detection.triggered_tip,
                PinTipDetection.INVALID_POSITION,
            )
            trigger = load_centre_collect_composite.pin_tip_detection.trigger
            trigger.return_value = NullStatus()  # type:ignore
            trigger.side_effect = None  # type: ignore

    RE.subscribe(wait_for_first_oav_grid)

    with pytest.raises(WarningException, match="No pin found after 5.0 seconds"):
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_params,
                oav_parameters_for_rotation,
            )
        )

    assert (
        fetch_blsample(load_centre_collect_params.sample_id).blSampleStatus
        == "ERROR - sample"
    )


@pytest.mark.system_test
def test_load_centre_collect_updates_bl_sample_status_gridscan_no_diffraction(
    composite_with_no_diffraction: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_blsample: Callable[..., Any],
):
    robot_load_cb = RobotLoadISPyBCallback()
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    sample_handling_cb = SampleHandlingCallback()
    RE.subscribe(robot_load_cb)
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(sample_handling_cb)

    with pytest.raises(CrystalNotFoundException):
        RE(
            load_centre_collect_full(
                composite_with_no_diffraction,
                load_centre_collect_params,
                oav_parameters_for_rotation,
            )
        )

    assert (
        fetch_blsample(load_centre_collect_params.sample_id).blSampleStatus
        == "ERROR - sample"
    )


@pytest.mark.system_test
def test_load_centre_collect_updates_bl_sample_status_rotation_failure(
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_blsample: Callable[..., Any],
):
    robot_load_cb = RobotLoadISPyBCallback()
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    sample_handling_cb = SampleHandlingCallback()
    RE.subscribe(robot_load_cb)
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(sample_handling_cb)

    with (
        patch(
            "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.arm_zebra",
            side_effect=TimeoutError("Simulated timeout"),
        ),
        pytest.raises(TimeoutError, match="Simulated timeout"),
    ):
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_params,
                oav_parameters_for_rotation,
            )
        )

    assert (
        fetch_blsample(load_centre_collect_params.sample_id).blSampleStatus
        == "ERROR - beamline"
    )


@pytest.mark.parametrize(
    "zocalo_result, expected_exception",
    [
        [TEST_RESULT_MEDIUM, nullcontext()],
        [TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX, nullcontext()],
        [TEST_RESULT_IN_BOUNDS_TOP_LEFT_GRID_CORNER, nullcontext()],
        [
            TEST_RESULT_OUT_OF_BOUNDS_COM,
            pytest.raises(IndexError, match=".* is outside the bounds of the grid"),
        ],
        [
            TEST_RESULT_OUT_OF_BOUNDS_BB,
            pytest.raises(IndexError, match=".* is outside the bounds of the grid"),
        ],
    ],
)
@pytest.mark.system_test
def test_load_centre_collect_gridscan_result_at_edge_of_grid(
    zocalo_result,
    expected_exception,
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    robot_load_cb: RobotLoadISPyBCallback,
    RE: RunEngine,
):
    load_centre_collect_composite.zocalo.my_zocalo_result = _with_sample_ids(
        zocalo_result, [SimConstants.ST_SAMPLE_ID]
    )
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    set_mock_value(
        load_centre_collect_composite.undulator_dcm.undulator_ref().current_gap, 1.11
    )
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(ispyb_rotation_cb)
    RE.subscribe(robot_load_cb)
    with expected_exception:
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_params,
                oav_parameters_for_rotation,
            )
        )


@pytest.mark.system_test
def test_execute_load_centre_collect_capture_rotation_snapshots(
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollectiongroup_attribute: Callable[..., Any],
    fetch_datacollection_ids_for_group_id: Callable[..., Any],
    fetch_blsample: Callable[[int], BLSample],
    tmp_path: Path,
):
    load_centre_collect_params.multi_rotation_scan.snapshot_directory = tmp_path

    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    snapshot_callback = BeamDrawingCallback(emit=ispyb_rotation_cb)
    set_mock_value(
        load_centre_collect_composite.undulator_dcm.undulator_ref().current_gap, 1.11
    )
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(snapshot_callback)
    RE(
        load_centre_collect_full(
            load_centre_collect_composite,
            load_centre_collect_params,
            oav_parameters_for_rotation,
        )
    )

    EXPECTED_SNAPSHOT_VALUES = {
        "xtalSnapshotFullPath1": f"regex:{tmp_path}/\\d{{8}}_oav_snapshot_0_with_beam_centre\\.png",
        "xtalSnapshotFullPath2": f"regex:{tmp_path}/\\d{{8}}_oav_snapshot_90_with_beam_centre\\.png",
        "xtalSnapshotFullPath3": f"regex:{tmp_path}/\\d{{8}}_oav_snapshot_180_with_beam_centre\\.png",
        "xtalSnapshotFullPath4": f"regex:{tmp_path}/\\d{{8}}_oav_snapshot_270_with_beam_centre\\.png",
    }

    rotation_dcg_id = ispyb_rotation_cb.ispyb_ids.data_collection_group_id
    rotation_dc_ids = fetch_datacollection_ids_for_group_id(rotation_dcg_id)
    compare_actual_and_expected(
        rotation_dc_ids[0],
        EXPECTED_SNAPSHOT_VALUES,
        fetch_datacollection_attribute,
    )
    compare_actual_and_expected(
        rotation_dc_ids[1],
        EXPECTED_SNAPSHOT_VALUES,
        fetch_datacollection_attribute,
    )

    for column in [
        "xtalSnapshotFullPath1",
        "xtalSnapshotFullPath2",
        "xtalSnapshotFullPath3",
        "xtalSnapshotFullPath4",
    ]:
        filename = fetch_datacollection_attribute(rotation_dc_ids[0], column)
        assert_images_pixelwise_equal(
            filename, "tests/test_data/test_images/generate_snapshot_output.png"
        )


def _with_sample_ids(zocalo_results: list[dict], sample_ids: Iterable[int]):
    copied_results = [zr.copy() for zr in zocalo_results]
    for result, sample_id in zip(copied_results, sample_ids, strict=False):
        result["sample_id"] = sample_id
    return copied_results


@pytest.mark.system_test
@pytest.mark.parametrize(
    "zocalo_result",
    [
        _with_sample_ids(
            TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX + TEST_RESULT_MEDIUM + TEST_RESULT_SMALL,
            [
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[1],
            ],
        )
    ],
)
def test_load_centre_collect_multisample_pin_reports_correct_sample_ids_in_ispyb_gridscan(
    zocalo_result,
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_msp_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    robot_load_cb: RobotLoadISPyBCallback,
    fetch_datacollectiongroup_attribute: Callable[..., Any],
    fetch_datacollection_attribute: Callable[..., Any],
):
    load_centre_collect_composite.zocalo.my_zocalo_result = zocalo_result
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    snapshot_cb = BeamDrawingCallback(emit=ispyb_rotation_cb)

    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(snapshot_cb)
    RE.subscribe(robot_load_cb)

    RE(
        load_centre_collect_full(
            load_centre_collect_composite,
            load_centre_collect_msp_params,
            oav_parameters_for_rotation,
        )
    )

    expected_sample_id = load_centre_collect_msp_params.sample_id

    compare_actual_and_expected(
        ispyb_gridscan_cb.ispyb_ids.data_collection_group_id,
        {"blSampleId": expected_sample_id},
        fetch_datacollectiongroup_attribute,
    )

    compare_actual_and_expected(
        ispyb_gridscan_cb.ispyb_ids.data_collection_ids[0],
        {"BLSAMPLEID": expected_sample_id},
        fetch_datacollection_attribute,
    )
    compare_actual_and_expected(
        ispyb_gridscan_cb.ispyb_ids.data_collection_ids[1],
        {"BLSAMPLEID": expected_sample_id},
        fetch_datacollection_attribute,
    )


@pytest.mark.system_test
@pytest.mark.parametrize(
    "zocalo_result",
    [
        _with_sample_ids(
            TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX + TEST_RESULT_MEDIUM + TEST_RESULT_SMALL,
            [
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[1],
            ],
        )
    ],
)
def test_load_centre_collect_multisample_pin_reports_correct_sample_ids_in_ispyb_rotation(
    zocalo_result,
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_msp_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    robot_load_cb: RobotLoadISPyBCallback,
    fetch_datacollectiongroup_attribute: Callable[..., Any],
    fetch_datacollection_attribute: Callable[..., Any],
    fetch_datacollection_ids_for_group_id: Callable[..., Any],
):
    load_centre_collect_composite.zocalo.my_zocalo_result = zocalo_result
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    snapshot_cb = BeamDrawingCallback(emit=ispyb_rotation_cb)
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(snapshot_cb)
    RE.subscribe(robot_load_cb)

    original_upsert_dcg = ispyb_rotation_cb.ispyb._upsert_data_collection_group
    captured_upsert_dcg_ids = []

    def intercept_upserts(conn, params):
        dcg_id = original_upsert_dcg(conn, params)
        nonlocal captured_upsert_dcg_ids
        if dcg_id not in captured_upsert_dcg_ids:
            captured_upsert_dcg_ids.append(dcg_id)
        return dcg_id

    with patch.object(
        ispyb_rotation_cb.ispyb,
        "_upsert_data_collection_group",
        side_effect=intercept_upserts,
    ):
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_msp_params,
                oav_parameters_for_rotation,
            )
        )

    assert len(captured_upsert_dcg_ids) == 2
    for dcg_id, expected_sample_id in zip(
        captured_upsert_dcg_ids, SimConstants.ST_MSP_SAMPLE_IDS, strict=True
    ):
        compare_actual_and_expected(
            dcg_id,
            {"blSampleId": expected_sample_id},
            fetch_datacollectiongroup_attribute,
        )
        dc_ids = fetch_datacollection_ids_for_group_id(dcg_id)
        compare_actual_and_expected(
            dc_ids[0],
            {"BLSAMPLEID": expected_sample_id},
            fetch_datacollection_attribute,
        )
        compare_actual_and_expected(
            dc_ids[1],
            {"BLSAMPLEID": expected_sample_id},
            fetch_datacollection_attribute,
        )


@pytest.mark.parametrize(
    "zocalo_result",
    [
        _with_sample_ids(
            TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX + TEST_RESULT_MEDIUM + TEST_RESULT_SMALL,
            [
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[1],
            ],
        )
    ],
)
@pytest.mark.system_test
def test_load_centre_collect_multisample_pin_reports_correct_sample_ids_robot_load(
    zocalo_result,
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_msp_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    robot_load_cb: RobotLoadISPyBCallback,
):
    load_centre_collect_composite.zocalo.my_zocalo_result = zocalo_result
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    snapshot_cb = BeamDrawingCallback(emit=ispyb_rotation_cb)
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(snapshot_cb)
    RE.subscribe(robot_load_cb)

    RE(
        load_centre_collect_full(
            load_centre_collect_composite,
            load_centre_collect_msp_params,
            oav_parameters_for_rotation,
        )
    )

    expected_sample_id = load_centre_collect_msp_params.sample_id
    expected_proposal, expected_visit = get_proposal_and_session_from_visit_string(
        load_centre_collect_msp_params.visit
    )
    robot_load_cb.expeye.start_robot_action.assert_called_once_with(  # type: ignore
        "LOAD",
        expected_proposal,
        expected_visit,
        expected_sample_id,
    )


@pytest.mark.parametrize(
    "zocalo_result",
    [
        _with_sample_ids(
            TEST_RESULT_IN_BOUNDS_TOP_LEFT_BOX + TEST_RESULT_MEDIUM + TEST_RESULT_SMALL,
            [
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[0],
                SimConstants.ST_MSP_SAMPLE_IDS[1],
            ],
        )
    ],
)
@pytest.mark.system_test
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan._move_and_rotation",
    new=MagicMock(side_effect=AssertionError("Simulated error in rotation")),
)
def test_load_centre_collect_multisample_pin_updates_sample_status_for_parent_sample_when_error_in_rotation_on_child_sample(
    zocalo_result,
    load_centre_collect_composite: LoadCentreCollectComposite,
    load_centre_collect_msp_params: LoadCentreCollect,
    oav_parameters_for_rotation: OAVParameters,
    RE: RunEngine,
    robot_load_cb: RobotLoadISPyBCallback,
    fetch_blsample: Callable[..., Any],
):
    load_centre_collect_composite.zocalo.my_zocalo_result = zocalo_result
    ispyb_gridscan_cb = GridscanISPyBCallback(
        param_type=GridCommonWithHyperionDetectorParams
    )
    ispyb_rotation_cb = RotationISPyBCallback()
    snapshot_cb = BeamDrawingCallback(emit=ispyb_rotation_cb)
    sample_handling_cb = SampleHandlingCallback()
    RE.subscribe(ispyb_gridscan_cb)
    RE.subscribe(snapshot_cb)
    RE.subscribe(robot_load_cb)
    RE.subscribe(sample_handling_cb)

    unpatched_move_and_rotation = _move_and_rotation
    num_calls = 0

    def throw_on_third_call_wrapper(plan, *args, **kwargs):
        nonlocal num_calls
        num_calls += 1
        if num_calls == 3:
            raise AssertionError("Simulated error in rotation")
        yield from plan(*args, **kwargs)

    with patch(
        "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan._move_and_rotation",
        partial(throw_on_third_call_wrapper, unpatched_move_and_rotation),
    ):
        with pytest.raises(AssertionError, match="Simulated error in rotation"):
            RE(
                load_centre_collect_full(
                    load_centre_collect_composite,
                    load_centre_collect_msp_params,
                    oav_parameters_for_rotation,
                )
            )

    assert (
        fetch_blsample(SimConstants.ST_MSP_SAMPLE_IDS[0]).blSampleStatus
        == "ERROR - beamline"
    )


@pytest.fixture
def patch_detect_grid_and_do_gridscan_with_detected_pin_position(
    load_centre_collect_composite: LoadCentreCollectComposite,
):
    wrapped = detect_grid_and_do_gridscan

    # Before we do the grid scan, pretend we detected the pin at this position and move to it
    # This is the base snapshot position
    def wrapper(*args, **kwargs):
        yield from bps.mv(
            load_centre_collect_composite.smargon.x,
            -0.614,
            load_centre_collect_composite.smargon.y,
            0.0259,
            load_centre_collect_composite.smargon.z,
            0.250,
        )

        yield from wrapped(*args, **kwargs)

    with patch(
        "mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan.detect_grid_and_do_gridscan",
    ) as patched_detect_grid:
        patched_detect_grid.side_effect = wrapper
        yield patched_detect_grid


@pytest.fixture
def grid_detect_for_snapshot_generation():
    fake_grid_params = GridParamUpdate(
        x_start_um=-598.4,
        y_start_um=-215.3,
        y2_start_um=-215.3,
        z_start_um=150.6,
        z2_start_um=150.6,
        x_steps=30,
        y_steps=20,
        z_steps=13,
        x_step_size_um=20,
        y_step_size_um=20,
        z_step_size_um=20,
    )
    with patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback"
    ) as gdc:
        gdc.return_value.get_grid_parameters.return_value = fake_grid_params
        yield fake_grid_params


class TestGenerateSnapshot:
    @pytest.fixture()
    def test_config_files(self):
        return {
            "zoom_params_file": "tests/test_data/test_jCameraManZoomLevels.xml",
            "oav_config_json": "tests/test_data/test_daq_configuration/OAVCentring_hyperion.json",
            "display_config": "tests/test_data/test_daq_configuration/display.configuration",
        }

    @pytest.mark.system_test
    def test_load_centre_collect_generate_rotation_snapshots(
        self,
        load_centre_collect_composite: LoadCentreCollectComposite,
        load_centre_collect_params: LoadCentreCollect,
        grid_detect_for_snapshot_generation: GridParamUpdate,
        patch_detect_grid_and_do_gridscan_with_detected_pin_position: MagicMock,
        next_oav_system_test_image: MagicMock,
        RE: RunEngine,
        tmp_path: Path,
        test_config_files: dict,
        fetch_datacollection_attribute: Callable[..., Any],
        fetch_datacollection_ids_for_group_id: Callable[..., Any],
    ):
        oav_parameters = OAVParameters(
            oav_config_json=test_config_files["oav_config_json"],
            context="xrayCentring",
        )
        next_fake_snapshot = iter(
            [
                # 1 extra for robot load
                "tests/test_data/test_images/thau_1_91_0.png",
                "tests/test_data/test_images/thau_1_91_90.png",
                "tests/test_data/test_images/thau_1_91_0.png",
            ]
        )

        next_oav_system_test_image.side_effect = lambda: next(next_fake_snapshot)

        load_centre_collect_params.multi_rotation_scan.snapshot_directory = tmp_path
        load_centre_collect_params.robot_load_then_centre.snapshot_directory = (
            tmp_path / "grid_snapshots"
        )
        os.mkdir(load_centre_collect_params.robot_load_then_centre.snapshot_directory)
        load_centre_collect_params.multi_rotation_scan.use_grid_snapshots = True
        load_centre_collect_params.multi_rotation_scan.snapshot_omegas_deg = None
        load_centre_collect_composite.zocalo.my_zocalo_result = (
            SNAPSHOT_GENERATION_ZOCALO_RESULT
        )

        ispyb_gridscan_cb = GridscanISPyBCallback(
            param_type=GridCommonWithHyperionDetectorParams
        )
        ispyb_rotation_cb = RotationISPyBCallback()
        snapshot_callback = BeamDrawingCallback(emit=ispyb_rotation_cb)
        RE.subscribe(ispyb_gridscan_cb)
        RE.subscribe(snapshot_callback)
        RE(
            load_centre_collect_full(
                load_centre_collect_composite,
                load_centre_collect_params,
                oav_parameters,
            )
        )

        EXPECTED_GRID_SNAPSHOT_VALUES_0 = {
            "xtalSnapshotFullPath1": f"regex:{tmp_path}/grid_snapshots/robot_load_centring_file_1_90_grid_overlay.png",
            "xtalSnapshotFullPath2": f"regex:{tmp_path}/grid_snapshots/robot_load_centring_file_1_90_outer_overlay.png",
            "xtalSnapshotFullPath3": f"regex:{tmp_path}/grid_snapshots/robot_load_centring_file_1_90.png",
        }
        EXPECTED_GRID_SNAPSHOT_VALUES_1 = {
            "xtalSnapshotFullPath1": f"regex:{tmp_path}/grid_snapshots/robot_load_centring_file_1_0_grid_overlay.png",
            "xtalSnapshotFullPath2": f"regex:{tmp_path}/grid_snapshots/robot_load_centring_file_1_0_outer_overlay.png",
            "xtalSnapshotFullPath3": f"regex:{tmp_path}/grid_snapshots/robot_load_centring_file_1_0.png",
        }
        grid_dcg_id = ispyb_gridscan_cb.ispyb_ids.data_collection_group_id
        grid_dc_ids = fetch_datacollection_ids_for_group_id(grid_dcg_id)
        compare_actual_and_expected(
            grid_dc_ids[0],
            EXPECTED_GRID_SNAPSHOT_VALUES_0,
            fetch_datacollection_attribute,
        )
        compare_actual_and_expected(
            grid_dc_ids[1],
            EXPECTED_GRID_SNAPSHOT_VALUES_1,
            fetch_datacollection_attribute,
        )

        EXPECTED_ROTATION_SNAPSHOT_VALUES = {
            "xtalSnapshotFullPath1": f"regex:{tmp_path}/\\d{{8}}_oav_snapshot_robot_load_centring_file_1_90\\.png",
            "xtalSnapshotFullPath2": f"regex:{tmp_path}/\\d{{8}}_oav_snapshot_robot_load_centring_file_1_0\\.png",
        }

        rotation_dcg_id = ispyb_rotation_cb.ispyb_ids.data_collection_group_id
        rotation_dc_ids = fetch_datacollection_ids_for_group_id(rotation_dcg_id)
        compare_actual_and_expected(
            rotation_dc_ids[0],
            EXPECTED_ROTATION_SNAPSHOT_VALUES,
            fetch_datacollection_attribute,
        )
        compare_actual_and_expected(
            rotation_dc_ids[1],
            EXPECTED_ROTATION_SNAPSHOT_VALUES,
            fetch_datacollection_attribute,
        )

        for expected_path, actual_path in zip(
            [
                "tests/test_data/test_images/thau_1_91_expected_270.png",
                "tests/test_data/test_images/thau_1_91_expected_270.png",
                "tests/test_data/test_images/thau_1_91_expected_0.png",
                "tests/test_data/test_images/thau_1_91_expected_0.png",
            ],
            [
                fetch_datacollection_attribute(rotation_dc_ids[i], col)
                for col in ["xtalSnapshotFullPath1", "xtalSnapshotFullPath2"]
                for i in (0, 1)
            ],
            strict=False,
        ):
            assert_images_pixelwise_equal(actual_path, expected_path)



from collections.abc import Generator
from typing import cast
from unittest.mock import ANY, MagicMock, call, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.oav.pin_image_recognition import PinTipDetection

from mx_bluesky.common.parameters.constants import PlanNameConstants
from mx_bluesky.hyperion.experiment_plans.hyperion_grid_detect_then_xray_centre_plan import (
    hyperion_grid_detect_then_xray_centre,
)
from mx_bluesky.hyperion.parameters.device_composites import (
    HyperionGridDetectThenXRayCentreComposite,
)
from mx_bluesky.hyperion.parameters.gridscan import (
    GridScanWithEdgeDetect,
    HyperionSpecifiedThreeDGridScan,
)
from tests.conftest import TEST_RESULT_LARGE, simulate_xrc_result
from tests.unit_tests.common.experiment_plans.test_common_flyscan_xray_centre_plan import (
    CompleteException,
)


class TestHyperionGridDetectThenXrayCentrePlan:
    def test_full_hyperion_grid_scan(
        self,
        hyperion_fgs_params: HyperionSpecifiedThreeDGridScan,
        test_config_files: dict[str, str],
    ):
        devices = MagicMock()
        plan = hyperion_grid_detect_then_xray_centre(
            devices,
            cast(GridScanWithEdgeDetect, hyperion_fgs_params),
            test_config_files["oav_config_json"],
        )
        assert isinstance(plan, Generator)

    @patch(
        "mx_bluesky.common.preprocessors.preprocessors.check_and_pause_feedback",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.preprocessors.preprocessors.unpause_xbpm_feedback_and_set_transmission_to_1",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
        autospec=True,
    )
    @patch("bluesky.plan_stubs.sleep", autospec=True)
    def test_flyscan_xray_centre_unpauses_xbpm_feedback_on_exception(
        self,
        mock_sleep: MagicMock,
        mock_grid_detection_callback: MagicMock,
        mock_create_parameters_for_flyscan_xray_centre: MagicMock,
        mock_grid_detection_plan: MagicMock,
        mock_common_flyscan_xray_centre: MagicMock,
        mock_unpause_and_set_transmission: MagicMock,
        mock_check_and_pause: MagicMock,
        hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
        test_full_grid_scan_params: GridScanWithEdgeDetect,
        test_config_files,
        pin_tip_detection_with_found_pin: PinTipDetection,
        RE: RunEngine,
    ):
        class TestException(Exception):
            pass

        mock_common_flyscan_xray_centre.side_effect = TestException
        with pytest.raises(TestException):  # noqa: B017
            RE(
                hyperion_grid_detect_then_xray_centre(
                    hyperion_grid_detect_xrc_devices,
                    test_full_grid_scan_params,
                    test_config_files["oav_config_json"],
                )
            )

        # Called once on exception and once on close_run
        mock_unpause_and_set_transmission.assert_has_calls([call(ANY, ANY)])

    @patch("bluesky.plan_stubs.sleep", autospec=True)
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.XRayCentreEventHandler"
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.change_aperture_then_move_to_xtal"
    )
    def test_hyperion_grid_detect_then_xray_centre_pauses_and_unpauses_xbpm_feedback_in_correct_order(
        self,
        mock_change_aperture_then_move: MagicMock,
        mock_events_handler: MagicMock,
        mock_create_parameters: MagicMock,
        mock_grid_detection_callback: MagicMock,
        mock_grid_detection_plan: MagicMock,
        mock_check_topup: MagicMock,
        mock_wait: MagicMock,
        sim_run_engine: RunEngineSimulator,
        test_full_grid_scan_params: GridScanWithEdgeDetect,
        hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
        test_config_files,
        hyperion_fgs_params,
    ):
        flyscan_event_handler = MagicMock()
        flyscan_event_handler.xray_centre_results = "dummy"
        mock_events_handler.return_value = flyscan_event_handler
        mock_create_parameters.return_value = hyperion_fgs_params
        simulate_xrc_result(
            sim_run_engine,
            hyperion_grid_detect_xrc_devices.zocalo,
            TEST_RESULT_LARGE,
        )

        msgs = sim_run_engine.simulate_plan(
            hyperion_grid_detect_then_xray_centre(
                hyperion_grid_detect_xrc_devices,
                test_full_grid_scan_params,
                test_config_files["oav_config_json"],
            ),
        )

        # Assert order: pause -> open run -> close run -> unpause (set attenuator)
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "trigger" and msg.obj.name == "xbpm_feedback",
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "open_run"
            and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
        )

        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "close_run"
            and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
        )

        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == "attenuator"
            and msg.args == (1.0,),
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.bps.abs_set",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.move_aperture_if_required",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.stage",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan._fetch_xrc_results_from_zocalo",
    )
    @patch(
        "dodal.plans.preprocessors.verify_undulator_gap.verify_undulator_gap",
    )
    def test_hyperion_grid_detect_then_xray_centre_does_undulator_check_before_collection(
        self,
        mock_verify_gap: MagicMock,
        mock_fetch_zocalo_results: MagicMock,
        mock_run_gridscan: MagicMock,
        mock_bps_stage: MagicMock,
        mock_create_parameters: MagicMock,
        mock_grid_params_callback: MagicMock,
        mock_move_aperture_if_required: MagicMock,
        mock_bps_abs_set: MagicMock,
        mock_grid_detection_plan: MagicMock,
        RE: RunEngine,
        test_full_grid_scan_params: GridScanWithEdgeDetect,
        hyperion_grid_detect_xrc_devices: HyperionGridDetectThenXRayCentreComposite,
        test_config_files,
        hyperion_fgs_params,
    ):
        mock_create_parameters.return_value = hyperion_fgs_params
        mock_run_gridscan.side_effect = CompleteException
        with pytest.raises(CompleteException):
            RE(
                hyperion_grid_detect_then_xray_centre(
                    hyperion_grid_detect_xrc_devices,
                    test_full_grid_scan_params,
                    test_config_files["oav_config_json"],
                )
            )

        mock_verify_gap.assert_called_once()



from asyncio import Event
from unittest.mock import MagicMock, patch

import bluesky.plan_stubs as bps
import pytest
from blueapi.core import BlueskyContext
from bluesky import RunEngine
from bluesky.utils import MsgGenerator

from mx_bluesky.common.parameters.constants import Actions, Status
from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.runner import Command, GDARunner

from .conftest import launch_test_in_runner_event_loop


@pytest.fixture
def context(RE: RunEngine) -> BlueskyContext:
    return MagicMock(run_engine=RE)


@pytest.fixture
def mock_composite():
    return MagicMock()


@pytest.fixture
def runner(context, mock_composite):
    with patch.dict(
        "mx_bluesky.hyperion.experiment_plans.experiment_registry.PLAN_REGISTRY",
        {
            "load_centre_collect_full": {
                "setup": mock_composite,
                "param_type": LoadCentreCollect,
            }
        },
    ):
        yield GDARunner(context)


def test_wait_on_queue_processes_start_command(
    runner: GDARunner, load_centre_collect_params: LoadCentreCollect, mock_composite
):
    mock_plan = MagicMock()
    runner.start(mock_plan, load_centre_collect_params, "load_centre_collect_full")
    runner._command_queue.put(Command(action=Actions.SHUTDOWN))
    runner.wait_on_queue()
    mock_plan.assert_called_once_with(
        mock_composite.return_value, load_centre_collect_params
    )
    assert runner.current_status.status == Status.IDLE.value


def test_wait_on_queue_intercepts_warning_exception_reports_failed_status(
    runner: GDARunner, load_centre_collect_params: LoadCentreCollect, mock_composite
):
    mock_plan = MagicMock(side_effect=WarningException("Mock warning"))
    runner.start(mock_plan, load_centre_collect_params, "load_centre_collect_full")
    runner._command_queue.put(Command(action=Actions.SHUTDOWN))
    runner.wait_on_queue()
    mock_plan.assert_called_once_with(
        mock_composite.return_value, load_centre_collect_params
    )
    assert runner.current_status.status == Status.FAILED.value


def test_wait_on_queue_intercepts_beamline_exception_reports_failed_status(
    runner: GDARunner, load_centre_collect_params: LoadCentreCollect, mock_composite
):
    mock_plan = MagicMock(side_effect=RuntimeError("Mock error"))
    runner.start(mock_plan, load_centre_collect_params, "load_centre_collect_full")
    runner._command_queue.put(Command(action=Actions.SHUTDOWN))
    runner.wait_on_queue()
    mock_plan.assert_called_once_with(
        mock_composite.return_value, load_centre_collect_params
    )
    assert runner.current_status.status == Status.FAILED.value


def test_wait_on_queue_stop_interrupts_running_plan(
    runner: GDARunner,
    load_centre_collect_params: LoadCentreCollect,
    mock_composite,
    executor,
):
    wait_for_plan_start = Event()

    def mock_plan(composite, params) -> MsgGenerator:
        wait_for_plan_start.set()
        yield from bps.sleep(10.0)

    async def wait_and_then_stop():
        await wait_for_plan_start.wait()
        runner.stop()

    runner.start(mock_plan, load_centre_collect_params, "load_centre_collect_full")
    runner._command_queue.put(Command(action=Actions.SHUTDOWN))
    stop_task = launch_test_in_runner_event_loop(wait_and_then_stop, runner, executor)
    runner.wait_on_queue()
    assert stop_task.done()



import os
from asyncio import run_coroutine_threadsafe, sleep
from concurrent.futures import Executor
from contextlib import nullcontext
from dataclasses import fields
from threading import Event
from typing import Any
from unittest.mock import ANY, MagicMock, call, patch

import pytest
from blueapi.core import BlueskyContext
from bluesky import Msg
from bluesky import plan_stubs as bps
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.baton import Baton
from dodal.utils import get_beamline_based_on_environment_variable
from ophyd_async.testing import get_mock_put, set_mock_value

from mx_bluesky.common.parameters.components import (
    PARAMETER_VERSION,
    MxBlueskyParameters,
)
from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.common.utils.context import (
    device_composite_from_context,
    find_device_in_context,
)
from mx_bluesky.common.utils.exceptions import SampleException, WarningException
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.baton_handler import (
    HYPERION_USER,
    NO_USER,
    _initialise_udc,
    run_forever,
    run_udc_when_requested,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
)
from mx_bluesky.hyperion.external_interaction.alerting.constants import Subjects
from mx_bluesky.hyperion.parameters.components import Wait
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.plan_runner import PlanException, PlanRunner
from mx_bluesky.hyperion.utils.context import setup_context

from .conftest import AGAMEMNON_WAIT_INSTRUCTION, launch_test_in_runner_event_loop

# For tests to complete reliably, these should all be successively much
# larger than each other

# Time to wait in test script between checks for a generic condition
SLEEP_FAST_SPIN_WAIT_S = 0.02
# Time for pytest to timeout if the script thread is deadlocked (shouldn't need this)
# PYTEST_TEST_TIMEOUT_S > TEST_SCRIPT_TIMEOUT in order that an exception on the script
# is bubbled up via the future and not lost.
PYTEST_TEST_TIMEOUT_S = 10


@pytest.fixture(autouse=True)
def patch_setup_devices(request):
    """
    Patch setup_devices() so that it doesn't rebuild all the devices
    if they already exist.
    """
    from mx_bluesky.hyperion.utils.context import setup_devices

    def patched_setup_devices(context: BlueskyContext, dev_mode: bool):
        if isinstance(context.run_engine, MagicMock):
            # We are using the sim run engine, device state doesn't matter
            setup_devices(context, True)
        else:
            try:
                find_device_in_context(context, "baton", Baton)
                # If we found a baton, leave devices alone and do nothing
            except ValueError:
                # No baton, setup devices and configure it
                setup_devices(context, True)
                baton_with_requested_user(context, HYPERION_USER)

    with (
        patch(
            "mx_bluesky.hyperion.baton_handler.setup_devices",
            side_effect=patched_setup_devices,
        ) as patched_func,
        nullcontext()
        if "dont_patch_clear_devices" in request.fixturenames
        else patch("mx_bluesky.hyperion.baton_handler.clear_all_device_caches"),
    ):
        yield patched_func


@pytest.fixture
def dont_patch_clear_devices():
    return


@pytest.fixture
def bluesky_context(
    RE: RunEngine,
    smargon,
    aperture_scatterguard,
    robot,
    lower_gonio,
    baton,
    use_beamline_t01,
):
    # Baton for real run engine

    # Set the initial baton state
    context = BlueskyContext(RE)

    def mock_load_module(module, **kwargs):
        for device in [smargon, aperture_scatterguard, robot, lower_gonio, baton]:
            context.register_device(device)

    context.with_dodal_module(
        get_beamline_based_on_environment_variable(),
        mock=True,
        fake_with_ophyd_sim=True,
    )

    baton_with_requested_user(context, HYPERION_USER)
    with patch.object(context, "with_dodal_module", mock_load_module):
        yield context


@pytest.fixture
def bluesky_context_with_sim_run_engine(sim_run_engine: RunEngineSimulator):
    baton_requested_user = HYPERION_USER

    # Baton for sim run engine
    def get_requested_user(msg):
        nonlocal baton_requested_user
        return {"readback": baton_requested_user}

    def set_requested_user(msg):
        nonlocal baton_requested_user
        baton_requested_user = msg.args[0]

    sim_run_engine.add_handler("locate", get_requested_user, "baton-requested_user")
    sim_run_engine.add_handler(
        "set",
        set_requested_user,  # type: ignore
        "baton-requested_user",
    )

    msgs = []

    def run_plan_in_sim(plan):
        msgs.extend(sim_run_engine.simulate_plan(plan))
        return sim_run_engine.return_value

    faked_run_engine = MagicMock(spec=RunEngine, side_effect=run_plan_in_sim)  # type: ignore

    # wait_for_connection in ensure_connected creates a bunch of awaitables
    # that will never be awaited by the simulator, let's not create them
    def dont_connect(*args, **kwargs):
        yield from bps.null()

    with (
        patch("blueapi.utils.connect_devices.ensure_connected", dont_connect),
        patch.dict(os.environ, {"BEAMLINE": "i03"}),
    ):
        context = BlueskyContext(faked_run_engine)
        context.with_dodal_module(
            get_beamline_based_on_environment_variable(),
            mock=True,
            fake_with_ophyd_sim=True,
        )
        yield msgs, context


@pytest.fixture
def single_collection_agamemnon_request(
    load_centre_collect_params, mock_load_centre_collect
):
    with (
        patch(
            "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
            side_effect=[[load_centre_collect_params], []],
        ),
        patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state"),
    ):
        yield mock_load_centre_collect


@pytest.fixture
def single_collection_agamemnon_request_then_wait_forever(
    load_centre_collect_params, mock_load_centre_collect
):
    with (
        patch(
            "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
            side_effect=[
                [load_centre_collect_params],
                [AGAMEMNON_WAIT_INSTRUCTION] * 1000,
            ],
        ),
        patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state"),
    ):
        yield mock_load_centre_collect


def baton_with_requested_user(
    bluesky_context: BlueskyContext, user: str = HYPERION_USER
) -> Baton:
    baton = find_device_in_context(bluesky_context, "baton", Baton)
    set_mock_value(baton.requested_user, user)
    return baton


@pytest.fixture()
def udc_runner(bluesky_context: BlueskyContext, RE: RunEngine) -> PlanRunner:
    return PlanRunner(bluesky_context, True)


@pytest.fixture
def mock_load_centre_collect():
    with (
        patch("mx_bluesky.hyperion.baton_handler.create_devices"),
        patch(
            "mx_bluesky.hyperion.baton_handler.load_centre_collect_full"
        ) as mock_plan,
    ):
        yield mock_plan


@pytest.fixture
def mock_create_params_from_agamemnon():
    with patch(
        "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
        return_value=[],
    ) as patched_func:
        yield patched_func


@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.bps.sleep")
@pytest.mark.timeout(10)
def test_loop_until_hyperion_requested(
    mock_sleep: MagicMock,
    udc_runner: PlanRunner,
    bluesky_context: BlueskyContext,
    mock_create_params_from_agamemnon: MagicMock,
):
    baton = baton_with_requested_user(bluesky_context, NO_USER)
    number_of_sleep_calls = 5

    def set_hyperion_requested(*args):
        yield from bps.null()
        set_mock_value(baton.requested_user, HYPERION_USER)

    mock_calls: list[Any] = [MagicMock()] * (number_of_sleep_calls - 1)
    mock_calls.append(set_hyperion_requested())
    mock_sleep.side_effect = mock_calls

    run_udc_when_requested(bluesky_context, udc_runner)

    assert mock_sleep.call_count == number_of_sleep_calls


@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
def test_when_hyperion_requested_then_hyperion_set_to_current_user(
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
    mock_create_params_from_agamemnon: MagicMock,
):
    baton = find_device_in_context(bluesky_context, "baton", Baton)

    run_udc_when_requested(bluesky_context, udc_runner)

    assert get_mock_put(baton.current_user).mock_calls[0] == call(
        HYPERION_USER, wait=True
    )


@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state")
def test_when_hyperion_requested_then_default_state_and_collection_run(
    default_state: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
    mock_create_params_from_agamemnon: MagicMock,
):
    parent_mock = MagicMock()
    parent_mock.attach_mock(default_state, "default_state")
    parent_mock.attach_mock(mock_create_params_from_agamemnon, "main_loop")

    run_udc_when_requested(bluesky_context, udc_runner)
    assert parent_mock.method_calls == [
        call.default_state(bluesky_context),
        call.main_loop(),
    ]


async def _assert_baton_released(baton: Baton):
    assert await baton.requested_user.get_value() != HYPERION_USER
    assert get_mock_put(baton.current_user).mock_calls[-1] == call(NO_USER, wait=True)


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_when_exception_raised_in_collection_then_loop_stops_and_baton_released(
    agamemnon: MagicMock,
    mock_load_centre_collect: MagicMock,
    bluesky_context: BlueskyContext,
    load_centre_collect_params: LoadCentreCollect,
    udc_runner: PlanRunner,
):
    mock_load_centre_collect.side_effect = ValueError()
    agamemnon.return_value = [load_centre_collect_params]

    with pytest.raises(PlanException) as e:
        run_udc_when_requested(bluesky_context, udc_runner)

    assert isinstance(e.value.__cause__, ValueError)
    baton = find_device_in_context(bluesky_context, "baton", Baton)
    assert mock_load_centre_collect.call_count == 1
    await _assert_baton_released(baton)


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_when_warning_exception_raised_in_collection_then_loop_continues(
    agamemnon: MagicMock,
    mock_load_centre_collect: MagicMock,
    bluesky_context: BlueskyContext,
    load_centre_collect_params: LoadCentreCollect,
    udc_runner: PlanRunner,
):
    mock_load_centre_collect.side_effect = [
        WarningException(),
        MagicMock(),
        ValueError(),
    ]
    agamemnon.return_value = [load_centre_collect_params]
    with pytest.raises(PlanException) as e:
        run_udc_when_requested(bluesky_context, udc_runner)

    assert isinstance(e.value.__cause__, ValueError)
    baton = find_device_in_context(bluesky_context, "baton", Baton)
    assert mock_load_centre_collect.call_count == 3
    await _assert_baton_released(baton)


@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state")
async def test_when_exception_raised_in_default_state_then_baton_released(
    default_state: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
):
    default_state.side_effect = [ValueError()]
    with pytest.raises(ValueError):
        run_udc_when_requested(bluesky_context, udc_runner)

    baton = find_device_in_context(bluesky_context, "baton", Baton)
    await _assert_baton_released(baton)


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_when_exception_raised_in_getting_agamemnon_instruction_then_loop_stops_and_baton_released(
    agamemnon: MagicMock,
    udc_runner: PlanRunner,
    bluesky_context: BlueskyContext,
    RE: RunEngine,
):
    agamemnon.side_effect = ValueError()
    with pytest.raises(ValueError):
        run_udc_when_requested(bluesky_context, udc_runner)

    baton = find_device_in_context(bluesky_context, "baton", Baton)
    await _assert_baton_released(baton)


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler.load_centre_collect_full")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_when_no_agamemnon_instructions_left_then_loop_stops_and_baton_released(
    collection: MagicMock,
    agamemnon: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
):
    agamemnon.return_value = None
    run_udc_when_requested(bluesky_context, udc_runner)

    baton = find_device_in_context(bluesky_context, "baton", Baton)
    collection.assert_not_called()
    await _assert_baton_released(baton)


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_when_other_user_requested_collection_finished_then_baton_released(
    agamemnon: MagicMock,
    bluesky_context: BlueskyContext,
    mock_load_centre_collect: MagicMock,
    load_centre_collect_params: LoadCentreCollect,
    udc_runner: PlanRunner,
    dont_patch_clear_devices,
):
    plan_continuing = MagicMock()
    agamemnon.return_value = [load_centre_collect_params]

    def fake_collection_with_baton_request_part_way_through(*args):
        baton = find_device_in_context(bluesky_context, "baton", Baton)
        yield from bps.null()
        yield from bps.mv(baton.requested_user, "OTHER_USER")
        plan_continuing()

    mock_load_centre_collect.side_effect = (
        fake_collection_with_baton_request_part_way_through
    )

    run_udc_when_requested(bluesky_context, udc_runner)
    baton = find_device_in_context(bluesky_context, "baton", Baton)
    mock_load_centre_collect.assert_called_once()
    plan_continuing.assert_called_once()
    await _assert_baton_released(baton)
    assert await baton.requested_user.get_value() == "OTHER_USER"


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state")
async def test_when_multiple_agamemnon_instructions_then_default_state_only_run_once(
    default_state: MagicMock,
    agamemnon: MagicMock,
    mock_load_centre_collect: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
):
    agamemnon.side_effect = [MagicMock(), MagicMock(), None]
    run_udc_when_requested(bluesky_context, udc_runner)
    default_state.assert_called_once()


@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_initialise_udc_reloads_all_devices(dont_patch_clear_devices):
    context = setup_context(True)
    devices_before_reset: LoadCentreCollectComposite = device_composite_from_context(
        context, LoadCentreCollectComposite
    )

    _initialise_udc(context, True)

    devices_after_reset: LoadCentreCollectComposite = device_composite_from_context(
        context, LoadCentreCollectComposite
    )

    for f in fields(devices_after_reset):
        device_after_reset = getattr(devices_after_reset, f.name)
        device_before_reset = getattr(devices_before_reset, f.name)
        assert device_before_reset is not device_after_reset, (
            f"{id(device_before_reset)} == {id(device_after_reset)}"
        )


@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    MagicMock(
        side_effect=[
            [
                Wait.model_validate(
                    {"duration_s": 12.34, "parameter_model_version": PARAMETER_VERSION}
                )
            ],
            [],
        ]
    ),
)
def test_baton_handler_loop_waits_if_wait_instruction_received(
    bluesky_context_with_sim_run_engine: tuple[list[Msg], BlueskyContext],
    sim_run_engine: RunEngineSimulator,
):
    msgs, context = bluesky_context_with_sim_run_engine
    udc_runner = PlanRunner(context, True)
    run_udc_when_requested(context, udc_runner)

    assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "sleep" and msg.args[0] == 12.34
    )


def test_main_loop_rejects_unrecognised_instruction_when_received(
    bluesky_context_with_sim_run_engine: tuple[list[Msg], BlueskyContext],
    sim_run_engine: RunEngineSimulator,
):
    msgs, context = bluesky_context_with_sim_run_engine
    with (
        pytest.raises(AssertionError, match="Unsupported instruction decoded"),
        patch(
            "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
            MagicMock(
                side_effect=[
                    [
                        MxBlueskyParameters.model_validate(
                            {"parameter_model_version": PARAMETER_VERSION}
                        )
                    ],
                    [],
                ]
            ),
        ),
    ):
        run_udc_when_requested(context, PlanRunner(context, True))


@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@pytest.mark.timeout(PYTEST_TEST_TIMEOUT_S)
async def test_shutdown_releases_the_baton(
    mock_create_params_from_agamemnon: MagicMock,
    udc_runner: PlanRunner,
    RE: RunEngine,
):
    mock_create_params_from_agamemnon.return_value = [
        Wait(duration_s=10, parameter_model_version=PARAMETER_VERSION)  # type: ignore
    ]

    async def wait_and_then_shutdown():
        while udc_runner.current_status != Status.BUSY:
            await sleep(0.1)
        udc_runner.shutdown()

    shutdown_task = run_coroutine_threadsafe(wait_and_then_shutdown(), RE.loop)

    run_forever(udc_runner)
    baton = find_device_in_context(udc_runner.context, "baton", Baton)
    await _assert_baton_released(baton)
    assert shutdown_task.done()


@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    side_effect=[
        [AGAMEMNON_WAIT_INSTRUCTION],
        [AGAMEMNON_WAIT_INSTRUCTION],
        AssertionError(
            "create_parameters_from_agamemnon was unexpectedly called a 3rd time"
        ),
    ],
)
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@pytest.mark.timeout(PYTEST_TEST_TIMEOUT_S)
async def test_run_forever_resumes_collection_when_baton_taken_away(
    mock_create_parameters_from_agamemnon: MagicMock,
    udc_runner: PlanRunner,
    executor: Executor,
):
    async def take_requested_baton_away_then_wait_for_release_then_re_request():
        while udc_runner.current_status != Status.BUSY:
            await sleep(SLEEP_FAST_SPIN_WAIT_S)
        try:
            baton = find_device_in_context(udc_runner.context, "baton", Baton)
            # un-request baton, hyperion should have processed first instruction
            await baton.requested_user.set(NO_USER)
            while await baton.current_user.get_value() != NO_USER:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
            assert len(mock_create_parameters_from_agamemnon.mock_calls) == 1
            # Re-request baton, wait until hyperion picks up baton
            await baton.requested_user.set(HYPERION_USER)
            while udc_runner.current_status != Status.BUSY:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
        finally:
            udc_runner.shutdown()

    future = launch_test_in_runner_event_loop(
        take_requested_baton_away_then_wait_for_release_then_re_request,
        udc_runner,
        executor,
    )
    run_forever(udc_runner)

    future.result()  # Ensure successful completion

    assert len(mock_create_parameters_from_agamemnon.mock_calls) == 2


@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    side_effect=[
        [AGAMEMNON_WAIT_INSTRUCTION],
        [],
        [AGAMEMNON_WAIT_INSTRUCTION],
    ],
)
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@pytest.mark.timeout(PYTEST_TEST_TIMEOUT_S)
async def test_run_forever_resumes_collection_when_normal_completion_and_baton_returned(
    mock_create_parameters_from_agamemnon: MagicMock,
    udc_runner: PlanRunner,
    executor: Executor,
):
    async def wait_for_baton_release_then_re_request():
        try:
            while udc_runner.current_status != Status.BUSY:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
            baton = find_device_in_context(udc_runner.context, "baton", Baton)
            while await baton.current_user.get_value() != NO_USER:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
            assert len(mock_create_parameters_from_agamemnon.mock_calls) == 2
            await baton.requested_user.set(HYPERION_USER)
            while udc_runner.current_status != Status.BUSY:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
        finally:
            udc_runner.shutdown()

    future = launch_test_in_runner_event_loop(
        wait_for_baton_release_then_re_request, udc_runner, executor
    )
    run_forever(udc_runner)

    future.result()  # Ensure successful completion
    assert len(mock_create_parameters_from_agamemnon.mock_calls) == 3


@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    side_effect=AssertionError("Runner started command processing without baton"),
)
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_run_forever_handles_shutdown_while_waiting_for_baton(
    mock_create_parameters_from_agamemnon: MagicMock,
    udc_runner: PlanRunner,
    executor: Executor,
):
    baton = find_device_in_context(udc_runner.context, "baton", Baton)
    await baton.requested_user.set(NO_USER)

    async def issue_shutdown_without_baton():
        try:
            await sleep(0.1)
            assert udc_runner.current_status == Status.IDLE
        finally:
            udc_runner.shutdown()

    future = launch_test_in_runner_event_loop(
        issue_shutdown_without_baton, udc_runner, executor
    )
    run_forever(udc_runner)
    future.result()  # Ensure successful completion


@pytest.mark.timeout(PYTEST_TEST_TIMEOUT_S)
@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    side_effect=[
        [AGAMEMNON_WAIT_INSTRUCTION],
        [AGAMEMNON_WAIT_INSTRUCTION],
        [],
    ],
)
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
def test_run_forever_clears_error_status_on_resume(
    mock_create_parameters_from_agamemnon: MagicMock,
    udc_runner: PlanRunner,
    executor: Executor,
):
    function_is_patched = Event()

    async def error_with_command_then_resume():
        try:
            with patch(
                "mx_bluesky.hyperion.baton_handler._runner_sleep",
                side_effect=RuntimeError("Simulated plan exception"),
            ):
                function_is_patched.set()
                while udc_runner.current_status != Status.FAILED:
                    await sleep(SLEEP_FAST_SPIN_WAIT_S)
            assert len(mock_create_parameters_from_agamemnon.mock_calls) == 1
            baton = find_device_in_context(udc_runner.context, "baton", Baton)
            while await baton.current_user.get_value() != NO_USER:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
            await baton.requested_user.set(HYPERION_USER)
            while udc_runner.current_status != Status.BUSY:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
        finally:
            udc_runner.shutdown()

    future = launch_test_in_runner_event_loop(
        error_with_command_then_resume, udc_runner, executor
    )
    function_is_patched.wait()
    run_forever(udc_runner)

    future.result()  # Ensure successful completion


@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    side_effect=[
        [AGAMEMNON_WAIT_INSTRUCTION],
        [],
    ],
)
@patch("mx_bluesky.hyperion.baton_handler.set_commissioning_signal")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
async def test_commissioning_signal_set_on_baton_acquire(
    mock_set_commissioning_signal: MagicMock,
    mock_create_parameters_from_agamemnon: MagicMock,
    udc_runner: PlanRunner,
    executor: Executor,
):
    baton = find_device_in_context(udc_runner.context, "baton", Baton)
    parent = MagicMock()
    parent.attach_mock(mock_set_commissioning_signal, "set_commissioning_signal")
    parent.attach_mock(get_mock_put(baton.current_user), "current_user")
    parent.attach_mock(
        mock_create_parameters_from_agamemnon, "create_parameters_from_agamemnon"
    )
    LOGGER.debug("Set no requested user")
    set_mock_value(baton.requested_user, NO_USER)

    async def release_baton_and_check_commissioning_signal_set():
        try:
            mock_set_commissioning_signal.assert_not_called()
            LOGGER.debug("Set requested user")
            set_mock_value(baton.requested_user, HYPERION_USER)
            LOGGER.debug("Wait for create_parameters call")
            while len(parent.create_parameters_from_agamemnon.mock_calls) == 0:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
            LOGGER.debug("Wait for idle")
            while udc_runner.current_status != Status.IDLE:
                await sleep(SLEEP_FAST_SPIN_WAIT_S)
            parent.assert_has_calls(
                [
                    call.current_user("Hyperion", wait=True),
                    call.set_commissioning_signal(baton.commissioning),
                    call.create_parameters_from_agamemnon(),
                    call.create_parameters_from_agamemnon(),
                ]
            )
        finally:
            LOGGER.debug("shutdown runner")
            udc_runner.shutdown()

    future = launch_test_in_runner_event_loop(
        release_baton_and_check_commissioning_signal_set, udc_runner, executor
    )
    run_forever(udc_runner)
    future.result()  # Ensure successful completion


@patch(
    "mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon",
    MagicMock(side_effect=[[AGAMEMNON_WAIT_INSTRUCTION], []]),
)
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.get_alerting_service")
def test_run_udc_when_requested_raises_udc_start_alert_when_baton_acquired(
    mock_get_alerting_service: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
):
    mock_alert_service = mock_get_alerting_service.return_value
    run_udc_when_requested(bluesky_context, udc_runner)

    assert mock_alert_service.raise_alert.mock_calls[0] == call(
        Subjects.UDC_STARTED, "Unattended Data Collection has started.", {}
    )


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.clear_all_device_caches", MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.get_alerting_service")
def test_run_udc_when_requested_raises_baton_release_udc_completed_event_when_hyperion_releases_the_baton(
    mock_get_alerting_service: MagicMock,
    mock_create_params: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
):
    parent = MagicMock()

    def wait_then_nothing():
        yield [AGAMEMNON_WAIT_INSTRUCTION]

        mock_alert_service = mock_get_alerting_service.return_value
        parent.attach_mock(mock_alert_service.raise_alert, "raise_alert")
        baton = find_device_in_context(bluesky_context, "baton", Baton)
        parent.attach_mock(get_mock_put(baton.current_user), "current_user")
        yield []

    mock_create_params.side_effect = wait_then_nothing()

    run_udc_when_requested(bluesky_context, udc_runner)

    parent.assert_has_calls(
        [
            call.raise_alert(
                Subjects.UDC_COMPLETED,
                "Hyperion UDC has completed all pending Agamemnon requests.",
                {},
            ),
            call.current_user(NO_USER, wait=True),
            call.raise_alert(
                Subjects.UDC_BATON_RELEASED,
                "Hyperion has released the baton. The baton is currently "
                "requested by: None",
                {},
            ),
        ]
    )


@patch("mx_bluesky.hyperion.baton_handler.create_parameters_from_agamemnon")
@patch("mx_bluesky.hyperion.baton_handler._move_to_udc_default_state", new=MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.clear_all_device_caches", MagicMock())
@patch("mx_bluesky.hyperion.baton_handler.get_alerting_service")
def test_run_udc_when_requested_raises_baton_release_event_when_baton_requested_from_hyperion(
    mock_get_alerting_service: MagicMock,
    mock_create_params: MagicMock,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
):
    parent = MagicMock()

    def wait_then_nothing():
        yield [AGAMEMNON_WAIT_INSTRUCTION]
        baton = find_device_in_context(bluesky_context, "baton", Baton)
        set_mock_value(baton.requested_user, "GDA")
        mock_alert_service = mock_get_alerting_service.return_value
        parent.attach_mock(mock_alert_service.raise_alert, "raise_alert")
        parent.attach_mock(get_mock_put(baton.current_user), "current_user")
        yield [AGAMEMNON_WAIT_INSTRUCTION]
        raise AssertionError(
            "This should never be reached as the baton has already been released."
        )

    mock_create_params.side_effect = wait_then_nothing()

    run_udc_when_requested(bluesky_context, udc_runner)

    parent.assert_has_calls(
        [
            call.current_user(NO_USER, wait=True),
            call.raise_alert(
                Subjects.UDC_BATON_RELEASED,
                "Hyperion has released the baton. The baton is currently "
                "requested by: GDA",
                {},
            ),
        ]
    )


@patch("mx_bluesky.hyperion.baton_handler.robot_unload")
def test_robot_unload_performed_when_no_more_agamemnon_instructions(
    mock_robot_unload,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
    single_collection_agamemnon_request,
    dont_patch_clear_devices,
):
    mock_load_centre_collect = single_collection_agamemnon_request
    mock_load_centre_collect.return_value = iter([])
    parent = MagicMock()
    parent.attach_mock(mock_load_centre_collect, "load_centre_collect_full")
    parent.attach_mock(mock_robot_unload, "robot_unload")

    run_udc_when_requested(bluesky_context, udc_runner)

    parent.assert_has_calls(
        [
            call.load_centre_collect_full(ANY, ANY),
            call.robot_unload(ANY, ANY, ANY, ANY, "cm31105-4"),
        ]
    )


@patch("mx_bluesky.hyperion.baton_handler.robot_unload")
def test_robot_unload_performed_when_baton_requested_away_from_hyperion(
    mock_robot_unload,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
    single_collection_agamemnon_request_then_wait_forever,
    dont_patch_clear_devices,
):
    def request_baton_away_from_hyperion(*args):
        baton = find_device_in_context(bluesky_context, "baton", Baton)
        yield from bps.abs_set(baton.requested_user, NO_USER)

    parent = MagicMock()
    mock_load_centre_collect = single_collection_agamemnon_request_then_wait_forever
    parent.attach_mock(mock_load_centre_collect, "load_centre_collect_full")
    parent.attach_mock(mock_robot_unload, "robot_unload")
    mock_load_centre_collect.side_effect = request_baton_away_from_hyperion

    run_udc_when_requested(bluesky_context, udc_runner)

    parent.assert_has_calls(
        [
            call.load_centre_collect_full(ANY, ANY),
            call.robot_unload(ANY, ANY, ANY, ANY, "cm31105-4"),
        ]
    )


@patch("mx_bluesky.hyperion.baton_handler.robot_unload")
def test_robot_unload_not_performed_when_beamline_error(
    mock_robot_unload,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
    single_collection_agamemnon_request,
):
    mock_load_centre_collect = single_collection_agamemnon_request
    mock_load_centre_collect.side_effect = RuntimeError("Simulated beamline error")

    with pytest.raises(PlanException):
        run_udc_when_requested(bluesky_context, udc_runner)

    mock_robot_unload.assert_not_called()


@patch("mx_bluesky.hyperion.baton_handler.robot_unload")
def test_robot_unload_still_performed_when_sample_exception(
    mock_robot_unload,
    bluesky_context: BlueskyContext,
    udc_runner: PlanRunner,
    single_collection_agamemnon_request,
    dont_patch_clear_devices,
):
    mock_load_centre_collect = single_collection_agamemnon_request
    parent = MagicMock()
    parent.attach_mock(mock_load_centre_collect, "load_centre_collect_full")
    parent.attach_mock(mock_robot_unload, "robot_unload")
    mock_load_centre_collect.side_effect = SampleException("Simulated beamline error")

    run_udc_when_requested(bluesky_context, udc_runner)

    parent.assert_has_calls(
        [
            call.load_centre_collect_full(ANY, ANY),
            call.robot_unload(ANY, ANY, ANY, ANY, "cm31105-4"),
        ]
    )



import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.beamlines import aithre
from dodal.devices.aithre_lasershaping.goniometer import Goniometer
from dodal.devices.aithre_lasershaping.laser_robot import LaserRobot

from mx_bluesky.beamlines.aithre_lasershaping import (
    go_to_zero,
    set_beamline_safe_on_robot,
)


@pytest.fixture
def goniometer(RE: RunEngine) -> Goniometer:
    return aithre.goniometer(connect_immediately=True, mock=True)


@pytest.fixture
def robot(RE: RunEngine) -> LaserRobot:
    return aithre.robot(connect_immediately=True, mock=True)


@pytest.mark.parametrize(
    "set_pv_name, set_value",
    [
        ["omega", -90],
        ["x", 0.5],
        ["y", -10],
        ["z", 2],
        ["sampy", 0.25],
        ["sampz", -0.76],
    ],
)
async def test_beamline_safe_reads_unsafe_correctly(
    sim_run_engine: RunEngineSimulator,
    robot: LaserRobot,
    goniometer: Goniometer,
    set_pv_name: str,
    set_value: float,
):
    # Should be replaced by read_handler when https://github.com/bluesky/bluesky/issues/1906 is fixed
    def locate_pv(_):
        return {"readback": set_value}

    sim_run_engine.add_handler("locate", locate_pv, "goniometer-" + set_pv_name)

    msgs = sim_run_engine.simulate_plan(set_beamline_safe_on_robot(robot, goniometer))
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "robot-set_beamline_safe"
        and msg.args[0] == "No",
    )


async def test_beamline_safe_reads_safe_correctly(
    sim_run_engine: RunEngineSimulator,
    robot: LaserRobot,
    goniometer: Goniometer,
):
    msgs = sim_run_engine.simulate_plan(set_beamline_safe_on_robot(robot, goniometer))

    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "robot-set_beamline_safe"
        and msg.args[0] == "On",
    )


@pytest.mark.parametrize("wait", [True, False])
async def test_go_to_zero_gives_expected_result(
    sim_run_engine: RunEngineSimulator, goniometer: Goniometer, wait: bool
):
    msgs = sim_run_engine.simulate_plan(go_to_zero(goniometer=goniometer, wait=wait))

    for name in ["omega", "x", "y", "z", "sampy", "sampz"]:
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"goniometer-{name}"
            and msg.args[0] == 0
            and msg.kwargs["group"] == "move_to_zero",
        )
    if wait:
        assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "wait" and msg.kwargs["group"] == "move_to_zero",
        )



import threading
from collections.abc import Callable

from blueapi.core import BlueskyContext
from bluesky.utils import MsgGenerator, RequestAbort

from mx_bluesky.common.parameters.constants import Status
from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.hyperion.runner import BaseRunner


class PlanException(Exception):
    """Identifies an exception that was encountered during plan execution."""

    pass


class PlanRunner(BaseRunner):
    """Runner that executes experiments from inside a running Bluesky plan"""

    def __init__(self, context: BlueskyContext, dev_mode: bool) -> None:
        super().__init__(context)
        self.current_status: Status = Status.IDLE
        self.is_dev_mode = dev_mode

    def execute_plan(
        self,
        experiment: Callable[[], MsgGenerator],
    ) -> MsgGenerator:
        """Execute the specified experiment plan.
        Args:
            experiment: The experiment to run
        Raises:
            PlanException: If the plan raised an exception
            RequestAbort: If the RunEngine aborted during execution"""

        self.current_status = Status.BUSY

        try:
            yield from experiment()
            self.current_status = Status.IDLE
        except WarningException as e:
            LOGGER.warning("Plan failed with warning", exc_info=e)
            self.current_status = Status.FAILED
        except RequestAbort:
            # This will occur when the run engine processes an abort when we shut down
            LOGGER.info("UDC Runner aborting")
            raise
        except Exception as e:
            LOGGER.error("Plan failed with exception", exc_info=e)
            self.current_status = Status.FAILED
            raise PlanException("Exception thrown in plan execution") from e

    def shutdown(self):
        """Performs a prompt shutdown. Aborts the run engine and terminates the loop
        waiting for messages."""

        def issue_abort():
            try:
                # abort() causes the run engine to throw a RequestAbort exception
                # inside the plan, which will propagate through the contingency wrappers.
                # When the plan returns, the run engine will raise RunEngineInterrupted
                self.RE.abort()
            except Exception as e:
                LOGGER.warning(
                    "Exception encountered when issuing abort() to RunEngine:",
                    exc_info=e,
                )

        LOGGER.info("Shutting down: Stopping the run engine gracefully")
        if self.current_status != Status.ABORTING:
            self.current_status = Status.ABORTING
            stopping_thread = threading.Thread(target=issue_abort)
            stopping_thread.start()
            return



from unittest.mock import MagicMock

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.collimation_table import CollimationTable
from dodal.devices.cryostream import CryoStream
from dodal.devices.cryostream import InOut as CryoInOut
from dodal.devices.fluorescence_detector_motion import FluorescenceDetector
from dodal.devices.fluorescence_detector_motion import InOut as FlouInOut
from dodal.devices.mx_phase1.beamstop import Beamstop, BeamstopPositions
from dodal.devices.scintillator import InOut, Scintillator
from dodal.testing import patch_all_motors
from ophyd_async.core import Signal, init_devices
from ophyd_async.epics.motor import Motor
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.experiment_plans.inner_plans.udc_default_state import (
    UDCDefaultDevices,
    move_to_udc_default_state,
)


@pytest.fixture
async def default_devices(aperture_scatterguard):
    async with init_devices(mock=True):
        cryo = CryoStream("")
        fluo = FluorescenceDetector("")
        beamstop = Beamstop("", MagicMock())
        scintillator = Scintillator("", MagicMock(), MagicMock(), name="scin")
        collimation_table = CollimationTable("")

    with (
        patch_all_motors(scintillator),
        patch_all_motors(collimation_table),
        patch_all_motors(beamstop),
    ):
        yield UDCDefaultDevices(
            cryo, fluo, beamstop, scintillator, aperture_scatterguard, collimation_table
        )


async def test_given_cryostream_temp_is_too_high_then_exception_raised(
    RE: RunEngine,
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    sim_run_engine.add_read_handler_for(
        default_devices.cryostream.temperature_k,
        default_devices.cryostream.MAX_TEMP_K + 10,
    )
    with pytest.raises(ValueError, match="temperature is too high"):
        sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))


async def test_given_cryostream_pressure_is_too_high_then_exception_raised(
    RE: RunEngine,
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    sim_run_engine.add_read_handler_for(
        default_devices.cryostream.back_pressure_bar,
        default_devices.cryostream.MAX_PRESSURE_BAR + 10,
    )
    with pytest.raises(ValueError, match="pressure is too high"):
        sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))


async def test_scintillator_is_moved_out_before_aperture_scatterguard_moved_in(
    RE: RunEngine,
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "scin-selected_pos"
        and msg.args[0] == InOut.OUT,
    )
    msgs = assert_message_and_return_remaining(msgs, lambda msg: msg.command == "wait")
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "aperture_scatterguard-selected_aperture"
        and msg.args[0] == ApertureValue.SMALL,
    )


def test_udc_default_state_runs_in_real_RE(
    RE: RunEngine, default_devices: UDCDefaultDevices
):
    set_mock_value(default_devices.cryostream.temperature_k, 100)
    set_mock_value(default_devices.cryostream.back_pressure_bar, 0.01)
    default_devices.scintillator._aperture_scatterguard().selected_aperture.get_value = MagicMock(
        return_value=ApertureValue.PARKED
    )

    RE(move_to_udc_default_state(default_devices))


def test_udc_default_state_group_contains_expected_items_and_is_waited_on(
    RE: RunEngine,
    sim_run_engine: RunEngineSimulator,
    default_devices: UDCDefaultDevices,
):
    msgs = sim_run_engine.simulate_plan(move_to_udc_default_state(default_devices))

    expected_group = "udc_default"

    def assert_expected_set(signal: Signal | Motor, value):
        return assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "set"
            and msg.obj.name == signal.name
            and msg.args[0] == value
            and msg.kwargs["group"] == expected_group,
        )

    msgs = assert_expected_set(
        default_devices.fluorescence_det_motion.pos, FlouInOut.OUT
    )
    coll = default_devices.collimation_table
    for device in [
        coll.inboard_y,
        coll.outboard_y,
        coll.upstream_y,
        coll.upstream_x,
        coll.downstream_x,
    ]:
        msgs = assert_expected_set(device, 0)

    msgs = assert_expected_set(
        default_devices.beamstop.selected_pos, BeamstopPositions.DATA_COLLECTION
    )

    msgs = assert_expected_set(
        default_devices.aperture_scatterguard.selected_aperture, ApertureValue.SMALL
    )

    msgs = assert_expected_set(default_devices.cryostream.course, CryoInOut.IN)
    msgs = assert_expected_set(default_devices.cryostream.fine, CryoInOut.IN)

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == expected_group,
    )



from __future__ import annotations

from dodal.devices.fast_grid_scan import (
    PandAGridScanParams,
    ZebraGridScanParams,
)

from mx_bluesky.common.parameters.gridscan import (
    GridCommon,
    SpecifiedThreeDGridScan,
)
from mx_bluesky.hyperion.external_interaction.config_server import (
    get_hyperion_config_client,
)


class GridCommonWithHyperionDetectorParams(GridCommon):
    """Used by models which require detector parameters but have no specifications of the grid"""

    # These detector params only exist so that we can properly select enable_dev_shm. Remove in
    # https://github.com/DiamondLightSource/hyperion/issues/1395"""
    @property
    def detector_params(self):
        params = super().detector_params
        params.enable_dev_shm = (
            get_hyperion_config_client().get_feature_flags().USE_GPU_RESULTS
        )
        return params


class HyperionSpecifiedThreeDGridScan(SpecifiedThreeDGridScan):
    """Hyperion's 3D grid scan deviates from the common class due to: optionally using a PandA, optionally using dev_shm for GPU analysis, and using a config server for features"""

    # These detector params only exist so that we can properly select enable_dev_shm. Remove in
    # https://github.com/DiamondLightSource/hyperion/issues/1395"""

    @property
    def detector_params(self):
        params = super().detector_params
        params.enable_dev_shm = (
            get_hyperion_config_client().get_feature_flags().USE_GPU_RESULTS
        )
        return params

    # Relative to common grid scan, stub offsets are defined by config server
    @property
    def FGS_params(self) -> ZebraGridScanParams:
        return ZebraGridScanParams(
            x_steps=self.x_steps,
            y_steps=self.y_steps,
            z_steps=self.z_steps,
            x_step_size_mm=self.x_step_size_um / 1000,
            y_step_size_mm=self.y_step_size_um / 1000,
            z_step_size_mm=self.z_step_size_um / 1000,
            x_start_mm=self.x_start_um / 1000,
            y1_start_mm=self.y_start_um / 1000,
            z1_start_mm=self.z_start_um / 1000,
            y2_start_mm=self.y2_start_um / 1000,
            z2_start_mm=self.z2_start_um / 1000,
            set_stub_offsets=get_hyperion_config_client()
            .get_feature_flags()
            .SET_STUB_OFFSETS,
            dwell_time_ms=self.exposure_time_s * 1000,
            transmission_fraction=self.transmission_frac,
        )

    @property
    def panda_FGS_params(self) -> PandAGridScanParams:
        if self.y_steps % 2 and self.z_steps > 0:
            # See https://github.com/DiamondLightSource/hyperion/issues/1118 for explanation
            raise OddYStepsException(
                "The number of Y steps must be even for a PandA gridscan"
            )
        return PandAGridScanParams(
            x_steps=self.x_steps,
            y_steps=self.y_steps,
            z_steps=self.z_steps,
            x_step_size_mm=self.x_step_size_um / 1000,
            y_step_size_mm=self.y_step_size_um / 1000,
            z_step_size_mm=self.z_step_size_um / 1000,
            x_start_mm=self.x_start_um / 1000,
            y1_start_mm=self.y_start_um / 1000,
            z1_start_mm=self.z_start_um / 1000,
            y2_start_mm=self.y2_start_um / 1000,
            z2_start_mm=self.z2_start_um / 1000,
            set_stub_offsets=get_hyperion_config_client()
            .get_feature_flags()
            .SET_STUB_OFFSETS,
            run_up_distance_mm=get_hyperion_config_client()
            .get_feature_flags()
            .PANDA_RUNUP_DISTANCE_MM,
            transmission_fraction=self.transmission_frac,
        )


class OddYStepsException(Exception): ...


class PinTipCentreThenXrayCentre(GridCommonWithHyperionDetectorParams):
    tip_offset_um: float = 0


class GridScanWithEdgeDetect(GridCommonWithHyperionDetectorParams):
    pass
